import numpy as np
import plotly.graph_objects as go

from qibocal.data import Data, DataUnits
from qibocal.plots.utils import get_color_state0, get_color_state1, get_data_subfolders


# Plot RO optimization with frequency
def ro_frequency(folder, routine, qubit, format):
    fig = go.Figure()

    # iterate over multiple data folders
    subfolder = get_data_subfolders(folder)[0]
    report_n = 0
    fitting_report = ""

    try:
        data = DataUnits.load_data(folder, subfolder, routine, format, "data")
        data.df = data.df[data.df["qubit"] == qubit]
    except:
        data = DataUnits(
            name="data",
            quantities={"frequency": "Hz", "delta_frequency": "Hz"},
            options=["iteration", "state"],
        )

    try:
        data_fit = Data.load_data(folder, subfolder, routine, format, "fit")
        data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]

    except:
        data_fit = Data(
            name="fit",
            quantities=[
                "frequency",
                "delta_frequency",
                "rotation_angle",
                "threshold",
                "fidelity",
                "assignment_fidelity",
                "average_state0",
                "average_state1",
            ],
        )

    # Plot raw results with sliders
    for frequency in data.df["delta_frequency"].unique():
        state0_data = data.df[
            (data.df["delta_frequency"] == frequency) & (data.df["state"] == 0)
        ]
        state1_data = data.df[
            (data.df["delta_frequency"] == frequency) & (data.df["state"] == 1)
        ]
        fit_data = data_fit.df[data_fit.df["delta_frequency"] == frequency.magnitude]
        fit_data["average_state0"] = data_fit.df["average_state0"].apply(
            lambda x: complex(x)
        )
        fit_data["average_state1"] = data_fit.df["average_state1"].apply(
            lambda x: complex(x)
        )

        # print(fit_data)
        fig.add_trace(
            go.Scatter(
                x=state0_data["i"].pint.to("V").pint.magnitude,
                y=state0_data["q"].pint.to("V").pint.magnitude,
                name=f"q{qubit}/r{report_n}: state 0",
                mode="markers",
                showlegend=True,
                opacity=0.7,
                marker=dict(size=3, color=get_color_state0(report_n)),
                visible=False,
            ),
        )

        fig.add_trace(
            go.Scatter(
                x=state1_data["i"].pint.to("V").pint.magnitude,
                y=state1_data["q"].pint.to("V").pint.magnitude,
                name=f"q{qubit}/r{report_n}: state 1",
                mode="markers",
                showlegend=True,
                opacity=0.7,
                marker=dict(size=3, color=get_color_state1(report_n)),
                visible=False,
            ),
        )

        fig.add_trace(
            go.Scatter(
                x=fit_data["average_state0"].apply(lambda x: np.real(x)).to_numpy(),
                y=fit_data["average_state0"].apply(lambda x: np.imag(x)).to_numpy(),
                name=f"q{qubit}/r{report_n}: mean state 0",
                showlegend=True,
                visible=False,
                mode="markers",
                marker=dict(size=10, color=get_color_state0(report_n)),
            ),
        )

        fig.add_trace(
            go.Scatter(
                x=fit_data["average_state1"].apply(lambda x: np.real(x)).to_numpy(),
                y=fit_data["average_state1"].apply(lambda x: np.imag(x)).to_numpy(),
                name=f"avg q{qubit}/r{report_n}: mean state 1",
                showlegend=True,
                visible=False,
                mode="markers",
                marker=dict(size=10, color=get_color_state1(report_n)),
            ),
        )

    # Show data for the first frequency
    for i in range(4):
        fig.data[i].visible = True

    # Add slider
    steps = []
    for i, freq in enumerate(data.df["frequency"].unique()):
        step = dict(
            method="update",
            args=[
                {"visible": [False] * len(fig.data)},
            ],
            label=f"{freq:.6f}",
        )
        for j in range(4):
            step["args"][0]["visible"][i * 4 + j] = True
        steps.append(step)

    sliders = [
        dict(
            currentvalue={"prefix": "frequency: "},
            steps=steps,
        )
    ]

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="i (V)",
        yaxis_title="q (V)",
        sliders=sliders,
        title=f"q{qubit}",
    )
    fig.update_yaxes(
        scaleanchor="x",
        scaleratio=1,
    )

    # Plot the fidelity as a function of frequency
    fig_fidelity = go.Figure()

    fig_fidelity.add_trace(
        go.Scatter(x=data_fit.df["frequency"], y=data_fit.df["assignment_fidelity"])
    )
    fig_fidelity.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="delta frequency (Hz)",
        yaxis_title="assignment fidelity (ratio)",
        title=f"q{qubit}",
    )
    # Add fitting report for the best fidelity
    fit_data = data_fit.df[data_fit.df["fidelity"] == data_fit.df["fidelity"].max()]
    title_text = f"q{qubit}/r{report_n} | average state 0: ({complex(fit_data['average_state0'].to_numpy()[0]):.6f})<br>"
    title_text += f"q{qubit}/r{report_n} | average state 1: ({complex(fit_data['average_state1'].to_numpy()[0]):.6f})<br>"
    title_text += f"q{qubit}/r{report_n} | rotation angle: {float(fit_data['rotation_angle'].to_numpy()[0]):.3f} | threshold = {float(fit_data['threshold'].to_numpy()[0]):.6f}<br>"
    title_text += f"q{qubit}/r{report_n} | fidelity: {float(fit_data['fidelity'].to_numpy()[0]):.3f}<br>"
    title_text += f"q{qubit}/r{report_n} | assignment fidelity: {float(fit_data['assignment_fidelity'].to_numpy()[0]):.3f}<br>"
    title_text += f"q{qubit}/r{report_n} | optimal frequency: {float(fit_data['frequency'].to_numpy()[0]):.3f} Hz<br><br>"
    fitting_report = fitting_report + title_text
    return [fig, fig_fidelity], fitting_report


# Plot RO optimization with amplitude
def ro_amplitude(folder, routine, qubit, format):
    fig = go.Figure()

    # iterate over multiple data folders
    subfolder = get_data_subfolders(folder)[0]
    report_n = 0
    fitting_report = ""

    try:
        data = DataUnits.load_data(folder, subfolder, routine, format, "data")
        data.df = data.df[data.df["qubit"] == qubit]
    except:
        data = DataUnits(
            name="data",
            quantities={
                "amplitude": "dimensionless",
                "delta_amplitude": "dimensionless",
            },
            options=["iteration", "state"],
        )

    try:
        data_fit = Data.load_data(folder, subfolder, routine, format, "fit")
        data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]

    except:
        data_fit = Data(
            name="fit",
            quantities=[
                "amplitude",
                "delta_amplitude",
                "rotation_angle",
                "threshold",
                "fidelity",
                "assignment_fidelity",
                "average_state0",
                "average_state1",
            ],
        )

    # Plot raw results with sliders
    for amplitude in data.df["delta_amplitude"].unique():
        state0_data = data.df[
            (data.df["delta_amplitude"] == amplitude) & (data.df["state"] == 0)
        ]
        state1_data = data.df[
            (data.df["delta_amplitude"] == amplitude) & (data.df["state"] == 1)
        ]
        fit_data = data_fit.df[data_fit.df["delta_amplitude"] == amplitude.magnitude]
        fit_data["average_state0"] = data_fit.df["average_state0"].apply(
            lambda x: complex(x)
        )
        fit_data["average_state1"] = data_fit.df["average_state1"].apply(
            lambda x: complex(x)
        )

        # print(fit_data)
        fig.add_trace(
            go.Scatter(
                x=state0_data["i"].pint.to("V").pint.magnitude,
                y=state0_data["q"].pint.to("V").pint.magnitude,
                name=f"q{qubit}/r{report_n}: state 0",
                mode="markers",
                showlegend=True,
                opacity=0.7,
                marker=dict(size=3, color=get_color_state0(report_n)),
                visible=False,
            ),
        )

        fig.add_trace(
            go.Scatter(
                x=state1_data["i"].pint.to("V").pint.magnitude,
                y=state1_data["q"].pint.to("V").pint.magnitude,
                name=f"q{qubit}/r{report_n}: state 1",
                mode="markers",
                showlegend=True,
                opacity=0.7,
                marker=dict(size=3, color=get_color_state1(report_n)),
                visible=False,
            ),
        )
        fig.add_trace(
            go.Scatter(
                x=fit_data["average_state0"].apply(lambda x: np.real(x)).to_numpy(),
                y=fit_data["average_state0"].apply(lambda x: np.imag(x)).to_numpy(),
                name=f"q{qubit}/r{report_n}: mean state 0",
                showlegend=True,
                visible=False,
                mode="markers",
                marker=dict(size=10, color=get_color_state0(report_n)),
            ),
        )

        fig.add_trace(
            go.Scatter(
                x=fit_data["average_state1"].apply(lambda x: np.real(x)).to_numpy(),
                y=fit_data["average_state1"].apply(lambda x: np.imag(x)).to_numpy(),
                name=f"avg q{qubit}/r{report_n}: mean state 1",
                showlegend=True,
                visible=False,
                mode="markers",
                marker=dict(size=10, color=get_color_state1(report_n)),
            ),
        )

    report_n += 1

    # Show data for the first amplitude
    for i in range(4):
        fig.data[i].visible = True

    # Add slider
    steps = []
    for i, amp in enumerate(data.df["amplitude"].unique()):
        step = dict(
            method="update",
            args=[
                {"visible": [False] * len(fig.data)},
            ],
            label=f"{amp.magnitude:.4f}",
        )
        for j in range(4):
            step["args"][0]["visible"][i * 4 + j] = True
        steps.append(step)

    sliders = [
        dict(
            currentvalue={"prefix": "amplitude: "},
            steps=steps,
        )
    ]

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="i (V)",
        yaxis_title="q (V)",
        sliders=sliders,
        title=f"q{qubit}",
    )
    fig.update_yaxes(
        scaleanchor="x",
        scaleratio=1,
    )

    # Plot the fidelity as a function of amplitude
    fig_fidelity = go.Figure()

    fig_fidelity.add_trace(
        go.Scatter(x=data_fit.df["amplitude"], y=data_fit.df["assignment_fidelity"])
    )
    fig_fidelity.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="delta amplitude (dimensionless)",
        yaxis_title="assignment fidelity (ratio)",
        title=f"q{qubit}",
    )
    # Add fitting report for the best fidelity
    fit_data = data_fit.df[data_fit.df["fidelity"] == data_fit.df["fidelity"].max()]
    title_text = f"q{qubit}/r{report_n} | average state 0: ({complex(fit_data['average_state0'].to_numpy()[0]):.6f})<br>"
    title_text += f"q{qubit}/r{report_n} | average state 1: ({complex(fit_data['average_state1'].to_numpy()[0]):.6f})<br>"
    title_text += f"q{qubit}/r{report_n} | rotation angle: {float(fit_data['rotation_angle'].to_numpy()[0]):.3f} | threshold = {float(fit_data['threshold'].to_numpy()[0]):.6f}<br>"
    title_text += f"q{qubit}/r{report_n} | fidelity: {float(fit_data['fidelity'].to_numpy()[0]):.3f}<br>"
    title_text += f"q{qubit}/r{report_n} | assignment fidelity: {float(fit_data['assignment_fidelity'].to_numpy()[0]):.3f}<br>"
    title_text += f"q{qubit}/r{report_n} | optimal amplitude: {float(fit_data['amplitude'].to_numpy()[0]):.3f}<br><br>"
    fitting_report = fitting_report + title_text
    return [fig, fig_fidelity], fitting_report


# Plot RO optimization with power
# Plot RO optimization with amplitude
def ro_power(folder, routine, qubit, format):
    fig = go.Figure()

    # iterate over multiple data folders
    subfolder = get_data_subfolders(folder)[0]
    report_n = 0
    fitting_report = ""

    try:
        data = DataUnits.load_data(folder, subfolder, routine, format, "data")
        data.df = data.df[data.df["qubit"] == qubit]
    except:
        data = DataUnits(
            name="data",
            quantities={
                "power": "dBm",
                "delta_power": "dBm",
            },
            options=["iteration", "state"],
        )

    try:
        data_fit = Data.load_data(folder, subfolder, routine, format, "fit")
        data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]

    except:
        data_fit = Data(
            name="fit",
            quantities=[
                "power",
                "delta_power",
                "rotation_angle",
                "threshold",
                "fidelity",
                "assignment_fidelity",
                "average_state0",
                "average_state1",
            ],
        )

    # Plot raw results with sliders
    for power in data.df["delta_power"].unique():
        state0_data = data.df[
            (data.df["delta_power"] == power) & (data.df["state"] == 0)
        ]
        state1_data = data.df[
            (data.df["delta_power"] == power) & (data.df["state"] == 1)
        ]
        fit_data = data_fit.df[data_fit.df["delta_power"] == power.magnitude]
        fit_data["average_state0"] = data_fit.df["average_state0"].apply(
            lambda x: complex(x)
        )
        fit_data["average_state1"] = data_fit.df["average_state1"].apply(
            lambda x: complex(x)
        )

        # print(fit_data)
        fig.add_trace(
            go.Scatter(
                x=state0_data["i"].pint.to("V").pint.magnitude,
                y=state0_data["q"].pint.to("V").pint.magnitude,
                name=f"q{qubit}/r{report_n}: state 0",
                mode="markers",
                showlegend=True,
                opacity=0.7,
                marker=dict(size=3, color=get_color_state0(report_n)),
                visible=False,
            ),
        )

        fig.add_trace(
            go.Scatter(
                x=state1_data["i"].pint.to("V").pint.magnitude,
                y=state1_data["q"].pint.to("V").pint.magnitude,
                name=f"q{qubit}/r{report_n}: state 1",
                mode="markers",
                showlegend=True,
                opacity=0.7,
                marker=dict(size=3, color=get_color_state1(report_n)),
                visible=False,
            ),
        )
        fig.add_trace(
            go.Scatter(
                x=fit_data["average_state0"].apply(lambda x: np.real(x)).to_numpy(),
                y=fit_data["average_state0"].apply(lambda x: np.imag(x)).to_numpy(),
                name=f"q{qubit}/r{report_n}: mean state 0",
                showlegend=True,
                visible=False,
                mode="markers",
                marker=dict(size=10, color=get_color_state0(report_n)),
            ),
        )

        fig.add_trace(
            go.Scatter(
                x=fit_data["average_state1"].apply(lambda x: np.real(x)).to_numpy(),
                y=fit_data["average_state1"].apply(lambda x: np.imag(x)).to_numpy(),
                name=f"avg q{qubit}/r{report_n}: mean state 1",
                showlegend=True,
                visible=False,
                mode="markers",
                marker=dict(size=10, color=get_color_state1(report_n)),
            ),
        )

    report_n += 1

    # Show data for the first power
    for i in range(4):
        fig.data[i].visible = True

    # Add slider
    steps = []
    for i, amp in enumerate(data.df["power"].unique()):
        step = dict(
            method="update",
            args=[
                {"visible": [False] * len(fig.data)},
            ],
            label=f"{amp.magnitude:.4f}",
        )
        for j in range(4):
            step["args"][0]["visible"][i * 4 + j] = True
        steps.append(step)

    sliders = [
        dict(
            currentvalue={"prefix": "power: "},
            steps=steps,
        )
    ]

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="i (V)",
        yaxis_title="q (V)",
        sliders=sliders,
        title=f"q{qubit}",
    )
    fig.update_yaxes(
        scaleanchor="x",
        scaleratio=1,
    )

    # Plot the fidelity as a function of power
    fig_fidelity = go.Figure()

    fig_fidelity.add_trace(
        go.Scatter(x=data_fit.df["power"], y=data_fit.df["assignment_fidelity"])
    )
    fig_fidelity.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="delta power (dBm)",
        yaxis_title="assignment fidelity (ratio)",
        title=f"q{qubit}",
    )
    # Add fitting report for the best fidelity
    fit_data = data_fit.df[data_fit.df["fidelity"] == data_fit.df["fidelity"].max()]
    title_text = f"q{qubit}/r{report_n} | average state 0: ({complex(fit_data['average_state0'].to_numpy()[0]):.6f})<br>"
    title_text += f"q{qubit}/r{report_n} | average state 1: ({complex(fit_data['average_state1'].to_numpy()[0]):.6f})<br>"
    title_text += f"q{qubit}/r{report_n} | rotation angle: {float(fit_data['rotation_angle'].to_numpy()[0]):.3f} | threshold = {float(fit_data['threshold'].to_numpy()[0]):.6f}<br>"
    title_text += f"q{qubit}/r{report_n} | fidelity: {float(fit_data['fidelity'].to_numpy()[0]):.3f}<br>"
    title_text += f"q{qubit}/r{report_n} | assignment fidelity: {float(fit_data['assignment_fidelity'].to_numpy()[0]):.3f}<br>"
    title_text += f"q{qubit}/r{report_n} | optimal power: {float(fit_data['power'].to_numpy()[0]):.3f} dBm<br><br>"
    fitting_report = fitting_report + title_text
    return [fig, fig_fidelity], fitting_report



import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal import plots
from qibocal.data import Data, DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import drag_tuning_fit

# allXY rotations
gatelist = [
    ["I", "I"],
    ["RX(pi)", "RX(pi)"],
    ["RY(pi)", "RY(pi)"],
    ["RX(pi)", "RY(pi)"],
    ["RY(pi)", "RX(pi)"],
    ["RX(pi/2)", "I"],
    ["RY(pi/2)", "I"],
    ["RX(pi/2)", "RY(pi/2)"],
    ["RY(pi/2)", "RX(pi/2)"],
    ["RX(pi/2)", "RY(pi)"],
    ["RY(pi/2)", "RX(pi)"],
    ["RX(pi)", "RY(pi/2)"],
    ["RY(pi)", "RX(pi/2)"],
    ["RX(pi/2)", "RX(pi)"],
    ["RX(pi)", "RX(pi/2)"],
    ["RY(pi/2)", "RY(pi)"],
    ["RY(pi)", "RY(pi/2)"],
    ["RX(pi)", "I"],
    ["RY(pi)", "I"],
    ["RX(pi/2)", "RX(pi/2)"],
    ["RY(pi/2)", "RY(pi/2)"],
]


@plot("Probability vs Gate Sequence", plots.allXY)
def allXY(
    platform: AbstractPlatform,
    qubits: dict,
    beta_param=None,
    software_averages=1,
    points=10,
):
    r"""
    The AllXY experiment is a simple test of the calibration of single qubit gatesThe qubit (initialized in the |0> state)
    is subjected to two back-to-back single-qubit gates and measured. In each round, we run 21 different gate pairs:
    ideally, the first 5 return the qubit to |0>, the next 12 drive it to superposition state, and the last 4 put the
    qubit in |1> state.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        beta_param (float): Drag pi pulse coefficient. If none, teh default shape defined in the runcard will be used.
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Difference between resonator signal voltage mesurement in volts from sequence 1 and 2
            - **i[V]**: Difference between resonator signal voltage mesurement for the component I in volts from sequence 1 and 2
            - **q[V]**: Difference between resonator signal voltage mesurement for the component Q in volts from sequence 1 and 2
            - **phase[rad]**: Difference between resonator signal phase mesurement in radians from sequence 1 and 2
            - **probability[dimensionless]**: Probability of being in |0> state
            - **gateNumber[dimensionless]**: Gate number applied from the list of gates
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages
    """
    # reload instrument settings from runcard
    platform.reload_settings()

    # create a Data object to store the results
    data = Data(
        name="data",
        quantities={"probability", "gateNumber", "qubit", "iteration"},
    )

    count = 0
    # repeat the experiment as many times as defined by software_averages
    for iteration in range(software_averages):
        gateNumber = 1
        # sweep the parameter
        for gateNumber, gates in enumerate(gatelist):
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data

            # create a sequence of pulses
            ro_pulses = {}
            sequence = PulseSequence()
            for qubit in qubits:
                sequence, ro_pulses[qubit] = _add_gate_pair_pulses_to_sequence(
                    platform, gates, qubit, beta_param, sequence
                )

            # execute the pulse sequence
            results = platform.execute_pulse_sequence(sequence)

            # retrieve the results for every qubit
            for ro_pulse in ro_pulses.values():
                z_proj = 2 * results[ro_pulse.serial].ground_state_probability - 1
                # store the results
                r = {
                    "probability": z_proj,
                    "gateNumber": gateNumber,
                    "beta_param": beta_param,
                    "qubit": ro_pulse.qubit,
                    "iteration": iteration,
                }
                data.add(r)
            count += 1
    # finally, save the remaining data
    yield data


@plot("Probability vs Gate Sequence", plots.allXY_drag_pulse_tuning)
def allXY_drag_pulse_tuning(
    platform: AbstractPlatform,
    qubits: dict,
    beta_start,
    beta_end,
    beta_step,
    software_averages=1,
    points=10,
):
    r"""
    The AllXY experiment is a simple test of the calibration of single qubit gatesThe qubit (initialized in the |0> state)
    is subjected to two back-to-back single-qubit gates and measured. In each round, we run 21 different gate pairs:
    ideally, the first 5 return the qubit to |0>, the next 12 drive it to superposition state, and the last 4 put the
    qubit in |1> state.

    The AllXY iteration method allows the user to execute iteratively the list of gates playing with the drag pulse shape
    in order to find the optimal drag pulse coefficient for pi pulses.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        beta_start (float): Initial drag pulse beta parameter
        beta_end (float): Maximum drag pulse beta parameter
        beta_step (float): Scan range step for the drag pulse beta parameter
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Difference between resonator signal voltage mesurement in volts from sequence 1 and 2
            - **i[V]**: Difference between resonator signal voltage mesurement for the component I in volts from sequence 1 and 2
            - **q[V]**: Difference between resonator signal voltage mesurement for the component Q in volts from sequence 1 and 2
            - **phase[rad]**: Difference between resonator signal phase mesurement in radians from sequence 1 and 2
            - **probability[dimensionless]**: Probability of being in |0> state
            - **gateNumber[dimensionless]**: Gate number applied from the list of gates
            - **beta_param[dimensionless]**: Beta paramter applied in the current execution
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

    """

    # reload instrument settings from runcard
    platform.reload_settings()

    data = Data(
        name="data",
        quantities={"probability", "gateNumber", "beta_param", "qubit", "iteration"},
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    for iteration in range(software_averages):
        # sweep the parameters
        for beta_param in np.arange(beta_start, beta_end, beta_step).round(4):
            gateNumber = 1
            for gates in gatelist:
                # save data as often as defined by points
                if count % points == 0 and count > 0:
                    # save data
                    yield data

                # create a sequence of pulses
                ro_pulses = {}
                sequence = PulseSequence()
                for qubit in qubits:
                    sequence, ro_pulses[qubit] = _add_gate_pair_pulses_to_sequence(
                        platform, gates, qubit, beta_param, sequence
                    )

                # execute the pulse sequence
                results = platform.execute_pulse_sequence(sequence)

                # retrieve the results for every qubit
                for ro_pulse in ro_pulses.values():
                    z_proj = 2 * results[ro_pulse.serial].ground_state_probability - 1
                    # store the results
                    r = {
                        "probability": z_proj,
                        "gateNumber": gateNumber,
                        "beta_param": beta_param,
                        "qubit": ro_pulse.qubit,
                        "iteration": iteration,
                    }
                    data.add(r)
                count += 1
                gateNumber += 1
    # finally, save the remaining data
    yield data


@plot("MSR vs beta parameter", plots.drag_pulse_tuning)
def drag_pulse_tuning(
    platform: AbstractPlatform,
    qubits: dict,
    beta_start,
    beta_end,
    beta_step,
    software_averages=1,
    points=10,
):
    r"""
    In this experiment, we apply two sequences in a given qubit: Rx(pi/2) - Ry(pi) and Ry(pi) - Rx(pi/2) for a range
    of different beta parameter values. After fitting, we obtain the best coefficient value for a pi pulse with drag shape.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        beta_start (float): Initial drag pulse beta parameter
        beta_end (float): Maximum drag pulse beta parameter
        beta_step (float): Scan range step for the drag pulse beta parameter
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Difference between resonator signal voltage mesurement in volts from sequence 1 and 2
            - **i[V]**: Difference between resonator signal voltage mesurement for the component I in volts from sequence 1 and 2
            - **q[V]**: Difference between resonator signal voltage mesurement for the component Q in volts from sequence 1 and 2
            - **phase[rad]**: Difference between resonator signal phase mesurement in radians from sequence 1 and 2
            - **beta_param[dimensionless]**: Optimal drag coefficient
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **optimal_beta_param**: Best drag pulse coefficent
            - **popt0**: offset
            - **popt1**: oscillation amplitude
            - **popt2**: period
            - **popt3**: phase
            - **qubit**: The qubit being tested
    """

    # reload instrument settings from runcard
    platform.reload_settings()

    # define the parameter to sweep and its range:
    # qubit drive DRAG pulse beta parameter
    beta_param_range = np.arange(beta_start, beta_end, beta_step).round(4)

    # create a DataUnits object to store the MSR, phase, i, q and the beta parameter
    data = DataUnits(
        name="data",
        quantities={"beta_param": "dimensionless"},
        options=["qubit", "iteration"],
    )

    count = 0
    # repeat the experiment as many times as defined by software_averages
    for iteration in range(software_averages):
        for beta_param in beta_param_range:
            # save data as often as defined by points
            if count % points == 0 and count > 0:
                # save data
                yield data
                # calculate and save fit
                yield drag_tuning_fit(
                    data,
                    x="beta_param[dimensionless]",
                    y="MSR[uV]",
                    qubits=qubits,
                    labels=["optimal_beta_param"],
                )

            # create two sequences of pulses
            # seq1: RX(pi/2) - RY(pi) - MZ
            # seq1: RY(pi/2) - RX(pi) - MZ

            ro_pulses = {}
            seq1 = PulseSequence()
            seq2 = PulseSequence()
            for qubit in qubits:
                # drag pulse RX(pi/2)
                RX90_drag_pulse = platform.create_RX90_drag_pulse(
                    qubit, start=0, beta=beta_param
                )
                # drag pulse RY(pi)
                RY_drag_pulse = platform.create_RX_drag_pulse(
                    qubit,
                    start=RX90_drag_pulse.finish,
                    relative_phase=+np.pi / 2,
                    beta=beta_param,
                )
                # drag pulse RY(pi/2)
                RY90_drag_pulse = platform.create_RX90_drag_pulse(
                    qubit, start=0, relative_phase=np.pi / 2, beta=beta_param
                )
                # drag pulse RX(pi)
                RX_drag_pulse = platform.create_RX_drag_pulse(
                    qubit, start=RY90_drag_pulse.finish, beta=beta_param
                )

                # RO pulse
                ro_pulses[qubit] = platform.create_qubit_readout_pulse(
                    qubit,
                    start=2
                    * RX90_drag_pulse.duration,  # assumes all single-qubit gates have same duration
                )
                # RX(pi/2) - RY(pi) - RO
                seq1.add(RX90_drag_pulse)
                seq1.add(RY_drag_pulse)
                seq1.add(ro_pulses[qubit])

                # RX(pi/2) - RY(pi) - RO
                seq2.add(RY90_drag_pulse)
                seq2.add(RX_drag_pulse)
                seq2.add(ro_pulses[qubit])

            # execute the pulse sequences
            result1 = platform.execute_pulse_sequence(seq1)
            result2 = platform.execute_pulse_sequence(seq2)

            # retrieve the results for every qubit
            for ro_pulse in ro_pulses.values():
                r1 = result1[ro_pulse.serial]
                r2 = result2[ro_pulse.serial]
                # store the results
                r = {
                    "MSR[V]": r1.measurement.mean() - r2.measurement.mean(),
                    "i[V]": r1.i.mean() - r2.i.mean(),
                    "q[V]": r1.q.mean() - r2.q.mean(),
                    "phase[rad]": r1.phase.mean() - r2.phase.mean(),
                    "beta_param[dimensionless]": beta_param,
                    "qubit": ro_pulse.qubit,
                    "iteration": iteration,
                }
                data.add(r)
            count += 1

    yield data
    yield drag_tuning_fit(
        data,
        x="beta_param[dimensionless]",
        y="MSR[uV]",
        qubits=qubits,
        labels=["optimal_beta_param"],
    )


def _add_gate_pair_pulses_to_sequence(
    platform: AbstractPlatform, gates, qubit, beta_param, sequence
):
    pulse_duration = platform.create_RX_pulse(qubit, start=0).duration
    # All gates have equal pulse duration

    sequenceDuration = 0
    pulse_start = 0

    for gate in gates:
        if gate == "I":
            # print("Transforming to sequence I gate")
            pass

        if gate == "RX(pi)":
            # print("Transforming to sequence RX(pi) gate")
            if beta_param == None:
                RX_pulse = platform.create_RX_pulse(
                    qubit,
                    start=pulse_start,
                )
            else:
                RX_pulse = platform.create_RX_drag_pulse(
                    qubit,
                    start=pulse_start,
                    beta=beta_param,
                )
            sequence.add(RX_pulse)

        if gate == "RX(pi/2)":
            # print("Transforming to sequence RX(pi/2) gate")
            if beta_param == None:
                RX90_pulse = platform.create_RX90_pulse(
                    qubit,
                    start=pulse_start,
                )
            else:
                RX90_pulse = platform.create_RX90_drag_pulse(
                    qubit,
                    start=pulse_start,
                    beta=beta_param,
                )
            sequence.add(RX90_pulse)

        if gate == "RY(pi)":
            # print("Transforming to sequence RY(pi) gate")
            if beta_param == None:
                RY_pulse = platform.create_RX_pulse(
                    qubit,
                    start=pulse_start,
                    relative_phase=np.pi / 2,
                )
            else:
                RY_pulse = platform.create_RX_drag_pulse(
                    qubit,
                    start=pulse_start,
                    relative_phase=np.pi / 2,
                    beta=beta_param,
                )
            sequence.add(RY_pulse)

        if gate == "RY(pi/2)":
            # print("Transforming to sequence RY(pi/2) gate")
            if beta_param == None:
                RY90_pulse = platform.create_RX90_pulse(
                    qubit,
                    start=pulse_start,
                    relative_phase=np.pi / 2,
                )
            else:
                RY90_pulse = platform.create_RX90_drag_pulse(
                    qubit,
                    start=pulse_start,
                    relative_phase=np.pi / 2,
                    beta=beta_param,
                )
            sequence.add(RY90_pulse)

        sequenceDuration = sequenceDuration + pulse_duration
        pulse_start = pulse_duration

    # RO pulse starting just after pair of gates
    ro_pulse = platform.create_qubit_readout_pulse(qubit, start=sequenceDuration + 4)
    sequence.add(ro_pulse)
    return sequence, ro_pulse



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from qibolab.sweeper import Parameter, Sweeper

from qibocal.auto.operation import Parameters, Qubits, Results, Routine
from qibocal.config import log
from qibocal.data import DataUnits

from .utils import find_min_msr, get_max_freq, get_points_with_max_freq, norm


@dataclass
class ResonatorPunchoutAttenuationParameters(Parameters):
    """ResonatorPunchoutAttenuation runcard inputs."""

    freq_width: int
    """Width for frequency sweep relative  to the readout frequency (Hz)."""
    freq_step: int
    """Frequency step for sweep (Hz)."""
    min_att: int
    """Attenuation minimum value (dB)."""
    max_att: int
    """Attenuation maximum value (dB)."""
    step_att: int
    """Attenuation step (dB)."""
    nshots: int
    """Number of shots."""
    relaxation_time: int
    """Relaxation time (ns)."""


@dataclass
class ResonatorPunchoutAttenuationResults(Results):
    """ResonatorPunchoutAttenation outputs."""

    readout_frequency: Dict[List[Tuple], str] = field(
        metadata=dict(update="readout_frequency")
    )
    """Readout frequency for each qubit."""
    readout_attenuation: Dict[List[Tuple], str] = field(
        metadata=dict(update="readout_attenuation")
    )
    """Readout attenuation for each qubit."""
    bare_frequency: Optional[Dict[List[Tuple], str]] = field(
        metadata=dict(update="bare_resonator_frequency")
    )
    """Bare resonator frequency for each qubit."""
    lp_max_att: Dict[List[Tuple], str]
    """Maximum attenuation at low power for each qubit."""
    lp_min_att: Dict[List[Tuple], str]
    """Minimum attenuation at low power for each qubit."""
    hp_max_att: Dict[List[Tuple], str]
    """Maximum attenuation at high power for each qubit."""
    hp_min_att: Dict[List[Tuple], str]
    """Minimum attenuation at high power for each qubit."""


class ResonatorPunchoutAttenuationData(DataUnits):
    """ResonatorPunchoutAttenuation data acquisition."""

    def __init__(self, resonator_type):
        super().__init__(
            "data",
            {"frequency": "Hz", "attenuation": "dB"},
            options=[
                "qubit",
            ],
        )
        self._resonator_type = resonator_type

    @property
    def resonator_type(self):
        """Type of resonator (2D or 3D)."""
        return self._resonator_type


def _acquisition(
    params: ResonatorPunchoutAttenuationParameters,
    platform: AbstractPlatform,
    qubits: Qubits,
) -> ResonatorPunchoutAttenuationData:
    """Data acquisition for Punchout over attenuation."""
    # create a sequence of pulses for the experiment:
    # MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()

    ro_pulses = {}
    for qubit in qubits:
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=0)
        sequence.add(ro_pulses[qubit])

    # define the parameters to sweep and their range:
    # resonator frequency
    delta_frequency_range = np.arange(
        -params.freq_width // 2, params.freq_width // 2, params.freq_step
    )
    freq_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        [ro_pulses[qubit] for qubit in qubits],
    )

    # attenuation
    attenuation_range = np.arange(params.min_att, params.max_att, params.step_att)
    att_sweeper = Sweeper(
        Parameter.attenuation, attenuation_range, qubits=list(qubits.values())
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include resonator frequency and attenuation
    data = ResonatorPunchoutAttenuationData(platform.resonator_type)

    results = platform.sweep(
        sequence,
        att_sweeper,
        freq_sweeper,
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
    )

    # retrieve the results for every qubit
    for qubit in qubits:
        # average msr, phase, i and q over the number of shots defined in the runcard
        result = results[ro_pulses[qubit].serial]
        att = np.repeat(attenuation_range, len(delta_frequency_range))
        # store the results
        freqs = np.array(
            len(attenuation_range)
            * list(delta_frequency_range + ro_pulses[qubit].frequency)
        ).flatten()
        r = {k: v.ravel() for k, v in result.raw.items()}
        r.update(
            {
                "frequency[Hz]": freqs,
                "attenuation[dB]": att,
                "qubit": len(freqs) * [qubit],
            }
        )
        data.add_data_from_dict(r)

        # # Temporary fixe to force to reset the attenuation to the original value in qblox
        # # sweeper method returning to orig value not working for attenuation
        # # After fitting punchout the reload_settings will be called automatically
        # platform.reload_settings()
        # save data
    return data


def _fit(
    data: ResonatorPunchoutAttenuationData, fit_type="attenuation"
) -> ResonatorPunchoutAttenuationResults:
    """Fit frequency and attenuation at high and low power for a given resonator."""

    qubits = data.df["qubit"].unique()

    freq_lp_dict = {}
    lp_max_att_dict = {}
    lp_min_att_dict = {}
    freq_hp_dict = {}
    hp_max_att_dict = {}
    hp_min_att_dict = {}
    ro_att_dict = {}

    for qubit in qubits:
        averaged_data = (
            data.df[data.df["qubit"] == qubit]
            .drop(columns=["i", "q", "qubit"])
            .groupby(["frequency", fit_type], as_index=False)
            .mean()
        )
        try:
            normalised_data = averaged_data.groupby([fit_type], as_index=False)[
                ["MSR"]
            ].transform(norm)

            averaged_data_updated = averaged_data.copy()
            averaged_data_updated.update(normalised_data["MSR"])

            min_points = find_min_msr(
                averaged_data_updated, data.resonator_type, fit_type
            )
            vfunc = np.vectorize(lambda x: x.magnitude)
            min_points = vfunc(min_points)

            max_x = np.amax(min_points[:, 0])
            min_x = np.amin(min_points[:, 0])
            middle_x = (max_x + min_x) / 2

            hp_points = min_points[min_points[:, 0] < middle_x]
            lp_points = min_points[min_points[:, 0] >= middle_x]

            freq_hp = get_max_freq(hp_points)
            freq_lp = get_max_freq(lp_points)

            point_hp_max, point_hp_min = get_points_with_max_freq(min_points, freq_hp)
            point_lp_max, point_lp_min = get_points_with_max_freq(min_points, freq_lp)

            freq_lp = point_lp_max[0]
            lp_max_att = point_lp_max[1]
            lp_min_att = point_lp_min[1]
            freq_hp = point_hp_max[0]
            hp_max_att = point_hp_max[1]
            hp_min_att = point_hp_min[1]
            ro_att = round((lp_max_att + lp_min_att) / 2)
            ro_att = ro_att + 1 if ro_att % 2 == 1 else ro_att

        except:
            log.warning("resonator_punchout_fit: the fitting was not succesful")
            freq_lp = 0.0
            freq_hp = 0.0
            ro_att = 0.0
            lp_max_att = 0.0
            lp_min_att = 0.0
            hp_max_att = 0.0
            hp_min_att = 0.0

        freq_lp_dict[qubit] = freq_lp / 1e9
        freq_hp_dict[qubit] = freq_hp / 1e9
        ro_att_dict[qubit] = ro_att
        lp_max_att_dict[qubit] = lp_max_att
        lp_min_att_dict[qubit] = lp_min_att
        hp_max_att_dict[qubit] = hp_max_att
        hp_min_att_dict[qubit] = hp_min_att
        log.warning(
            f"max att: {lp_max_att} -  min att: {lp_min_att} -  readout_attenuation: {ro_att}"
        )

    return ResonatorPunchoutAttenuationResults(
        freq_lp_dict,
        ro_att_dict,
        freq_hp_dict,
        lp_max_att_dict,
        lp_min_att_dict,
        hp_max_att_dict,
        hp_min_att_dict,
    )


def _plot(
    data: ResonatorPunchoutAttenuationData,
    fit: ResonatorPunchoutAttenuationResults,
    qubit,
):
    """Plotting for ResonatorPunchoutAttenuation."""

    figures = []
    fitting_report = ""
    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
        subplot_titles=(
            "Normalised MSR",
            "phase (rad)",
        ),
    )

    # TODO: remove this function
    def norm(x):
        x_mags = x.pint.to("V").pint.magnitude
        return (x_mags - np.min(x_mags)) / (np.max(x_mags) - np.min(x_mags))

    qubit_data = data.df[data.df["qubit"] == qubit].drop(columns=["qubit"])

    normalised_data = qubit_data.groupby(["attenuation"], as_index=False)[
        ["MSR"]
    ].transform(norm)

    fig.add_trace(
        go.Heatmap(
            x=qubit_data["frequency"].pint.to("Hz").pint.magnitude,
            y=qubit_data["attenuation"].pint.to("dB").pint.magnitude,
            z=normalised_data["MSR"],
            colorbar_x=0.46,
        ),
        row=1,
        col=1,
    )
    fig.update_xaxes(title_text=f"{qubit}: Frequency (Hz)", row=1, col=1)
    fig.update_yaxes(title_text="Attenuation", row=1, col=1)
    fig.add_trace(
        go.Heatmap(
            x=qubit_data["frequency"].pint.to("Hz").pint.magnitude,
            y=qubit_data["attenuation"].pint.to("dB").pint.magnitude,
            z=qubit_data["phase"].pint.to("rad").pint.magnitude,
            colorbar_x=1.01,
        ),
        row=1,
        col=2,
    )
    fig.update_xaxes(title_text=f"{qubit}/: Frequency (Hz)", row=1, col=2)
    fig.update_yaxes(title_text="Attenuation", row=1, col=2)

    if len(data) > 0:
        fig.add_trace(
            go.Scatter(
                x=[
                    fit.readout_frequency[qubit] * 1e9,
                    fit.readout_frequency[qubit] * 1e9,
                    fit.bare_frequency[qubit] * 1e9,
                    fit.bare_frequency[qubit] * 1e9,
                ],
                y=[
                    fit.lp_max_att[qubit],
                    fit.lp_min_att[qubit],
                    fit.hp_max_att[qubit],
                    fit.hp_min_att[qubit],
                ],
                mode="markers",
                marker=dict(
                    size=8,
                    color="gray",
                    symbol="circle",
                ),
            )
        )
        title_text = ""
        title_text += f"{qubit} | Resonator Frequency at Low Power:  {fit.readout_frequency[qubit] * 1e9} Hz.<br>"
        title_text += f"{qubit} | Low Power Attenuation Range: {fit.lp_max_att[qubit]} - {fit.lp_min_att[qubit]} db.<br>"
        title_text += f"{qubit} | Resonator Frequency at High Power: {fit.bare_frequency[qubit] * 1e9} Hz.<br>"
        title_text += f"{qubit} | High Power Attenuation Range: {fit.hp_max_att[qubit]} - {fit.hp_min_att[qubit]} db.<br>"

        fitting_report = fitting_report + title_text
    else:
        fitting_report = "No fitting data"

    fig.update_layout(
        showlegend=False,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
    )

    figures.append(fig)

    return figures, fitting_report


resonator_punchout_attenuation = Routine(_acquisition, _fit, _plot)
"""ResonatorPunchoutAttenuation Routine object."""



from typing import Optional

import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from qibolab.sweeper import Parameter, Sweeper

from qibocal import plots
from qibocal.data import DataUnits
from qibocal.decorators import plot
from qibocal.fitting.methods import lorentzian_fit


@plot("MSR and Phase vs Qubit Drive Frequency", plots.frequency_msr_phase)
def qubit_spectroscopy(
    platform: AbstractPlatform,
    qubits: dict,
    freq_width: int,
    freq_step: int,
    drive_duration: int,
    drive_amplitude: Optional[float] = None,
    nshots: int = 1024,
    relaxation_time: int = 50,
    software_averages: int = 1,
):
    r"""
    Perform spectroscopy on the qubit.
    This routine executes a fast scan around the expected qubit frequency indicated in the platform runcard.
    Afterthat, a final sweep with more precision is executed centered in the new qubit frequency found.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        freq_width (int): Width frequency in HZ to perform the high resolution sweep
        freq_step (int): Step frequency in HZ for the high resolution sweep
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - Two DataUnits objects with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Qubit drive frequency value in Hz
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **qubit**: The qubit being tested
            - **drive_frequency**: frequency
            - **peak_voltage**: peak voltage
            - **popt0**: Lorentzian's amplitude
            - **popt1**: Lorentzian's center
            - **popt2**: Lorentzian's sigma
            - **popt3**: Lorentzian's offset
    """

    # reload instrument settings from runcard
    platform.reload_settings()
    # create a sequence of pulses for the experiment:
    # long drive probing pulse - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(
            qubit, start=0, duration=drive_duration
        )
        if drive_amplitude is not None:
            qd_pulses[qubit].amplitude = drive_amplitude
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    delta_frequency_range = np.arange(-freq_width // 2, freq_width // 2, freq_step)
    sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        pulses=[qd_pulses[qubit] for qubit in qubits],
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include qubit frequency
    data = DataUnits(
        name="data",
        quantities={"frequency": "Hz"},
        options=["qubit", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    for iteration in range(software_averages):
        results = platform.sweep(
            sequence, sweeper, nshots=nshots, relaxation_time=relaxation_time
        )

        # retrieve the results for every qubit
        for qubit, ro_pulse in ro_pulses.items():
            # average msr, phase, i and q over the number of shots defined in the runcard
            result = results[ro_pulse.serial]
            r = result.raw
            # store the results
            r.update(
                {
                    "frequency[Hz]": delta_frequency_range + qd_pulses[qubit].frequency,
                    "qubit": len(delta_frequency_range) * [qubit],
                    "iteration": len(delta_frequency_range) * [iteration],
                }
            )
            data.add_data_from_dict(r)

        # finally, save the remaining data and fits
        yield data
        yield lorentzian_fit(
            data,
            x="frequency[GHz]",
            y="MSR[uV]",
            qubits=qubits,
            resonator_type=platform.resonator_type,
            labels=["drive_frequency", "peak_voltage"],
        )


@plot(
    "MSR and Phase vs Qubit Drive Frequency and Flux Current",
    plots.frequency_flux_msr_phase,
)
def qubit_spectroscopy_flux(
    platform: AbstractPlatform,
    qubits: dict,
    drive_amplitude,
    freq_width,
    freq_step,
    bias_width,
    bias_step,
    fluxlines,
    nshots=1024,
    relaxation_time=50,
    software_averages=1,
):
    r"""
    Perform spectroscopy on the qubit modifying the bias applied in the flux control line.
    This routine works for multiqubit devices flux controlled.

    Args:
        platform (AbstractPlatform): Qibolab platform object
        qubits (dict): Dict of target Qubit objects to perform the action
        freq_width (int): Width frequency in HZ to perform the spectroscopy sweep
        freq_step (int): Step frequency in HZ for the spectroscopy sweep
        bias_width (float): Width bias in A for the flux bias sweep
        bias_step (float): Step bias in A for the flux bias sweep
        fluxlines (list): List of flux lines to use to perform the experiment. If it is set to "qubits", it uses each of
                        flux lines associated with the target qubits.
        software_averages (int): Number of executions of the routine for averaging results
        points (int): Save data results in a file every number of points

    Returns:
        - A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **frequency[Hz]**: Qubit drive frequency value in Hz
            - **bias[V]**: Current value in A applied to the flux line
            - **qubit**: The qubit being tested
            - **fluxline**: The fluxline being tested
            - **iteration**: The iteration number of the many determined by software_averages

        - A DataUnits object with the fitted data obtained with the following keys

            - **drive_frequency**: frequency
            - **peak_voltage**: peak voltage
            - **popt0**: Lorentzian's amplitude
            - **popt1**: Lorentzian's center
            - **popt2**: Lorentzian's sigma
            - **popt3**: Lorentzian's offset
            - **qubit**: The qubit being tested
    """
    # reload instrument settings from runcard
    platform.reload_settings()

    # create a sequence of pulses for the experiment:
    # long drive probing pulse - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    sequence = PulseSequence()
    ro_pulses = {}
    qd_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_qubit_drive_pulse(
            qubit, start=0, duration=2000
        )
        qd_pulses[qubit].amplitude = drive_amplitude
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    # qubit drive frequency
    delta_frequency_range = np.arange(-freq_width // 2, freq_width // 2, freq_step)
    frequency_sweeper = Sweeper(
        Parameter.frequency,
        delta_frequency_range,
        pulses=[qd_pulses[qubit] for qubit in qubits],
    )

    if fluxlines == "qubits":
        fluxlines = qubits

    # flux bias
    delta_bias_range = np.arange(-bias_width / 2, bias_width / 2, bias_step)
    bias_sweeper = Sweeper(
        Parameter.bias,
        delta_bias_range,
        qubits=fluxlines,
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include qubit frequency and flux bias
    data = DataUnits(
        name=f"data",
        quantities={"frequency": "Hz", "bias": "V"},
        options=["qubit", "fluxline", "iteration"],
    )

    # repeat the experiment as many times as defined by software_averages
    count = 0
    for iteration in range(software_averages):
        results = platform.sweep(
            sequence,
            bias_sweeper,
            frequency_sweeper,
            nshots=nshots,
            relaxation_time=relaxation_time,
        )

        # retrieve the results for every qubit
        for qubit, fluxline in zip(qubits, fluxlines):
            # average msr, phase, i and q over the number of shots defined in the runcard
            result = results[ro_pulses[qubit].serial]
            # store the results
            biases = np.repeat(
                delta_bias_range, len(delta_frequency_range)
            ) + platform.get_bias(fluxline)
            freqs = np.array(
                len(delta_bias_range)
                * list(delta_frequency_range + qd_pulses[qubit].frequency)
            ).flatten()
            r = {k: v.ravel() for k, v in result.raw.items()}
            r.update(
                {
                    "frequency[Hz]": freqs,
                    "bias[V]": biases,
                    "qubit": len(freqs) * [qubit],
                    "fluxline": len(freqs) * [fluxline],
                    "iteration": len(freqs) * [iteration],
                }
            )
            data.add_data_from_dict(r)

        # finally, save the remaining data and fits
        yield data



from dataclasses import dataclass, field
from typing import Dict, List, Tuple

import numpy as np
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from qibolab.sweeper import Parameter, Sweeper
from scipy.optimize import curve_fit

from qibocal.auto.operation import Parameters, Qubits, Results, Routine
from qibocal.config import log
from qibocal.data import DataUnits

from . import utils


@dataclass
class RabiAmplitudeParameters(Parameters):
    """RabiAmplitude runcard inputs."""

    min_amp_factor: float
    """Minimum amplitude multiplicative factor."""
    max_amp_factor: float
    """Maximum amplitude multiplicative factor."""
    step_amp_factor: float
    """Step amplitude multiplicative factor."""
    pulse_length: float
    """RX pulse duration (ns)."""
    nshots: int
    """Number of shots."""
    relaxation_time: float
    """Relaxation time (ns)."""


@dataclass
class RabiAmplitudeResults(Results):
    """RabiAmplitude outputs."""

    amplitude: Dict[List[Tuple], str] = field(metadata=dict(update="drive_amplitude"))
    """Drive amplitude for each qubit."""
    length: Dict[List[Tuple], str] = field(metadata=dict(update="drive_length"))
    """Drive pulse duration. Same for all qubits."""
    fitted_parameters: Dict[List[Tuple], List]
    """Raw fitted parameters."""


class RabiAmplitudeData(DataUnits):
    """RabiAmplitude data acquisition."""

    def __init__(self):
        super().__init__(
            "data",
            {"amplitude": "dimensionless", "length": "ns"},
            options=["qubit"],
        )


def _acquisition(
    params: RabiAmplitudeParameters, platform: AbstractPlatform, qubits: Qubits
) -> RabiAmplitudeData:
    r"""
    Data acquisition for Rabi experiment sweeping amplitude.
    In the Rabi experiment we apply a pulse at the frequency of the qubit and scan the drive pulse amplitude
    to find the drive pulse amplitude that creates a rotation of a desired angle.
    """

    # create a sequence of pulses for the experiment
    sequence = PulseSequence()
    qd_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        qd_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        qd_pulses[qubit].duration = params.pulse_length
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=qd_pulses[qubit].finish
        )
        sequence.add(qd_pulses[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    # qubit drive pulse amplitude
    qd_pulse_amplitude_range = np.arange(
        params.min_amp_factor,
        params.max_amp_factor,
        params.step_amp_factor,
    )
    sweeper = Sweeper(
        Parameter.amplitude,
        qd_pulse_amplitude_range,
        [qd_pulses[qubit] for qubit in qubits],
    )

    # create a DataUnits object to store the results,
    # DataUnits stores by default MSR, phase, i, q
    # additionally include qubit drive pulse amplitude
    data = RabiAmplitudeData()

    # sweep the parameter
    results = platform.sweep(
        sequence,
        sweeper,
        nshots=params.nshots,
        relaxation_time=params.relaxation_time,
    )
    for qubit in qubits:
        # average msr, phase, i and q over the number of shots defined in the runcard
        result = results[ro_pulses[qubit].serial]
        r = result.raw
        r.update(
            {
                "amplitude[dimensionless]": qd_pulses[qubit].amplitude
                * qd_pulse_amplitude_range,
                "length[ns]": len(qd_pulse_amplitude_range)
                * [qd_pulses[qubit].duration],
                "qubit": len(qd_pulse_amplitude_range) * [qubit],
            }
        )
        data.add_data_from_dict(r)
    return data


def _fit(data: RabiAmplitudeData) -> RabiAmplitudeResults:
    """Post-processing for RabiAmplitude."""
    qubits = data.df["qubit"].unique()

    pi_pulse_amplitudes = {}
    fitted_parameters = {}
    durations = {}

    for qubit in qubits:
        qubit_data = data.df[data.df["qubit"] == qubit]

        rabi_parameter = qubit_data["amplitude"].pint.to("dimensionless").pint.magnitude
        voltages = qubit_data["MSR"].pint.to("uV").pint.magnitude
        durations[qubit] = qubit_data["length"].pint.to("ns").pint.magnitude.unique()

        y_min = np.min(voltages.values)
        y_max = np.max(voltages.values)
        x_min = np.min(rabi_parameter.values)
        x_max = np.max(rabi_parameter.values)
        x = (rabi_parameter.values - x_min) / (x_max - x_min)
        y = (voltages.values - y_min) / (y_max - y_min)

        # Guessing period using fourier transform
        ft = np.fft.rfft(y)
        mags = abs(ft)
        index = np.argmax(mags) if np.argmax(mags) != 0 else np.argmax(mags[1:]) + 1
        f = x[index] / (x[1] - x[0])

        pguess = [0.5, 1, f, np.pi / 2]
        try:
            popt, _ = curve_fit(
                utils.rabi_amplitude_fit,
                x,
                y,
                p0=pguess,
                maxfev=100000,
                bounds=(
                    [0, 0, 0, -np.pi],
                    [1, 1, np.inf, np.pi],
                ),
            )
            translated_popt = [
                y_min + (y_max - y_min) * popt[0],
                (y_max - y_min) * popt[1],
                popt[2] / (x_max - x_min),
                popt[3] - 2 * np.pi * x_min / (x_max - x_min) * popt[2],
            ]
            pi_pulse_parameter = np.abs((1.0 / translated_popt[2]) / 2)

        except:
            log.warning("rabi_fit: the fitting was not succesful")
            pi_pulse_parameter = 0
            fitted_parameters = [0] * 4

        pi_pulse_amplitudes[qubit] = pi_pulse_parameter
        fitted_parameters[qubit] = translated_popt

    return RabiAmplitudeResults(pi_pulse_amplitudes, durations, fitted_parameters)


def _plot(data: RabiAmplitudeData, fit: RabiAmplitudeResults, qubit):
    """Plotting function for RabiAmplitude."""
    return utils.plot(data, fit, qubit)


rabi_amplitude = Routine(_acquisition, _fit, _plot)
"""RabiAmplitude Routine object."""



import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.data import Data, DataUnits
from qibocal.fitting.utils import exp
from qibocal.plots.utils import get_color, get_data_subfolders, load_data


# Spin echos
def spin_echo_time_msr(folder, routine, qubit, format):
    """Spin echo plotting routine:
    The routine plots the results of a modified Ramsey sequence with an additional Rx(pi) pulse placed symmetrically between the two Rx(pi/2) pulses.
    An exponential fit to this data gives a spin echo decay time T2.
    Args:
        folder (string): Folder name where the data and fitted data is located
        routine (string): Name of the calibration routine that calls this plotting method
        qubit (int): Target qubit to characterize
        format (string): Data file format. Supported formats are .csv and .pkl
    """
    figures = []

    fig = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=("MSR (V)",),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    fitting_report = ""
    for subfolder in subfolders:
        try:
            data = load_data(folder, subfolder, routine, format, "data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(
                name=f"data",
                quantities={"wait": "ns"},
                options=["qubit", "iteration"],
            )

        try:
            data_fit = load_data(folder, subfolder, routine, format, "fits")
            data_fit.df = data_fit.df[data_fit.df["qubit"] == qubit]
        except:
            data_fit = Data(
                quantities=[
                    "popt0",
                    "popt1",
                    "popt2",
                    "label1",
                ]
            )

        data.df = data.df.drop(columns=["i", "q", "phase", "qubit"])
        iterations = data.df["iteration"].unique()
        waits = data.df["wait"].unique()

        if len(iterations) > 1:
            opacity = 0.3
        else:
            opacity = 1
        for iteration in iterations:
            iteration_data = data.df[data.df["iteration"] == iteration]
            fig.add_trace(
                go.Scatter(
                    x=iteration_data["wait"],
                    y=iteration_data["MSR"] * 1e6,
                    marker_color=get_color(report_n),
                    opacity=opacity,
                    name=f"q{qubit}/r{report_n}",
                    showlegend=not bool(iteration),
                    legendgroup=f"q{qubit}/r{report_n}",
                ),
                row=1,
                col=1,
            )

        if len(iterations) > 1:
            data.df = data.df.drop(columns=["iteration"])  # pylint: disable=E1101
            fig.add_trace(
                go.Scatter(
                    x=waits.tolist(),
                    y=data.df.groupby("wait")["MSR"].mean()  # pylint: disable=E1101
                    * 1e6,
                    marker_color=get_color(report_n),
                    name=f"q{qubit}/r{report_n}: Average",
                    showlegend=True,
                    legendgroup=f"q{qubit}/r{report_n}: Average",
                ),
                row=1,
                col=1,
            )

        # add fitting trace
        if len(data) > 0 and (qubit in data_fit.df["qubit"].values):
            waitrange = np.linspace(
                min(data.df["wait"]),
                max(data.df["wait"]),
                2 * len(data),
            )
            params = data_fit.df[data_fit.df["qubit"] == qubit].to_dict(
                orient="records"
            )[0]

            fig.add_trace(
                go.Scatter(
                    x=waitrange,
                    y=exp(
                        waitrange,
                        float(data_fit.df["popt0"]),
                        float(data_fit.df["popt1"]),
                        float(data_fit.df["popt2"]),
                    ),
                    name=f"q{qubit}/r{report_n} Fit",
                    line=go.scatter.Line(dash="dot"),
                    marker_color=get_color(4 * report_n + 2),
                ),
                row=1,
                col=1,
            )

            fitting_report = fitting_report + (
                f"q{qubit}/r{report_n} | t2: {params['t2']:,.0f} ns.<br><br>"
            )
        report_n += 1

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Time (ns)",
        yaxis_title="MSR (uV)",
    )

    figures.append(fig)

    return figures, fitting_report



from dataclasses import dataclass, field
from typing import Dict, List, Tuple

import numpy as np
import plotly.graph_objects as go
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence
from scipy.optimize import curve_fit

from qibocal.auto.operation import Parameters, Qubits, Results, Routine
from qibocal.config import log
from qibocal.data import DataUnits
from qibocal.plots.utils import get_color


@dataclass
class FlippingParameters(Parameters):
    """Flipping runcard inputs."""

    nflips_max: int
    """Maximum number of flips ([RX(pi) - RX(pi)] sequences). """
    nflips_step: int
    """Flip step."""


@dataclass
class FlippingResults(Results):
    """Flipping outputs."""

    amplitude: Dict[List[Tuple], str] = field(metadata=dict(update="drive amplitude"))
    """Drive amplitude for each qubit."""
    amplitude_factors: Dict[List[Tuple], str]
    """Drive amplitude correction factor for each qubit."""
    fitted_parameters: Dict[List[Tuple], List]
    """Raw fitting output."""


class FlippngData(DataUnits):
    """Flipping acquisition outputs."""

    def __init__(self, resonator_type):
        super().__init__(
            name="data",
            quantities={"flips": "dimensionless"},
            options=["qubit"],
        )

        self._resonator_type = resonator_type

    @property
    def resonator_type(self):
        """Type of resonator (2D or 3D)."""
        return self._resonator_type


def _acquisition(
    params: FlippingParameters,
    platform: AbstractPlatform,
    qubits: Qubits,
) -> FlippngData:
    r"""
    Data acquisition for flipping.

    The flipping experiment correct the delta amplitude in the qubit drive pulse. We measure a qubit after applying
    a Rx(pi/2) and N flips (Rx(pi) rotations). After fitting we can obtain the delta amplitude to refine pi pulses.

    Args:
        params (:class:`SingleShotClassificationParameters`): input parameters
        platform (:class:`AbstractPlatform`): Qibolab's platform
        qubits (dict): Dict of target :class:`Qubit` objects to be characterized

    Returns:
        data (:class:`FlippingData`)
    """

    # create a DataUnits object to store MSR, phase, i, q and the number of flips
    data = FlippngData(platform.resonator_type)

    # sweep the parameter
    for flips in range(0, params.nflips_max, params.nflips_step):
        # create a sequence of pulses for the experiment
        sequence = PulseSequence()
        ro_pulses = {}
        for qubit in qubits:
            RX90_pulse = platform.create_RX90_pulse(qubit, start=0)
            sequence.add(RX90_pulse)
            # execute sequence RX(pi/2) - [RX(pi) - RX(pi)] from 0...flips times - RO
            start1 = RX90_pulse.duration
            for j in range(flips):
                RX_pulse1 = platform.create_RX_pulse(qubit, start=start1)
                start2 = start1 + RX_pulse1.duration
                RX_pulse2 = platform.create_RX_pulse(qubit, start=start2)
                sequence.add(RX_pulse1)
                sequence.add(RX_pulse2)
                start1 = start2 + RX_pulse2.duration

            # add ro pulse at the end of the sequence
            ro_pulses[qubit] = platform.create_qubit_readout_pulse(qubit, start=start1)
            sequence.add(ro_pulses[qubit])
        # execute the pulse sequence
        results = platform.execute_pulse_sequence(sequence)
        for ro_pulse in ro_pulses.values():
            # average msr, phase, i and q over the number of shots defined in the runcard

            r = results[ro_pulse.serial].average.raw
            r.update(
                {
                    "flips[dimensionless]": flips,
                    "qubit": ro_pulse.qubit,
                    "pi_pulse_amplitude": qubits[ro_pulse.qubit].pi_pulse_amplitude,
                }
            )

            data.add(r)

    return data


def flipping_fit(x, p0, p1, p2, p3):
    # A fit to Flipping Qubit oscillation
    # Epsilon?? should be Amplitude : p[0]
    # Offset                        : p[1]
    # Period of oscillation         : p[2]
    # phase for the first point corresponding to pi/2 rotation   : p[3]
    return np.sin(x * 2 * np.pi / p2 + p3) * p0 + p1


# FIXME: not working
def _fit(data: FlippngData) -> FlippingResults:
    r"""Post-processing function for Flipping.

    The used model is

    .. math::

        y = p_0 sin\Big(\frac{2 \pi x}{p_2} + p_3\Big) + p_1.
    """
    qubits = data.df["qubit"].unique()
    corrected_amplitudes = {}
    fitted_parameters = {}
    amplitude_correction_factors = {}
    for qubit in qubits:
        qubit_data_df = data.df[data.df["qubit"] == qubit]
        pi_pulse_amplitude = qubit_data_df["pi_pulse_amplitude"].unique()
        voltages = qubit_data_df["MSR"].pint.to("uV").pint.magnitude
        flips = qubit_data_df["flips"].pint.magnitude

        if data.resonator_type == "3D":
            pguess = [
                pi_pulse_amplitude / 2,
                np.mean(voltages),
                -40,
                0,
            ]  # epsilon guess parameter
        else:
            pguess = [
                pi_pulse_amplitude / 2,
                np.mean(voltages),
                40,
                0,
            ]  # epsilon guess parameter

        try:
            popt, _ = curve_fit(flipping, flips, voltages, p0=pguess, maxfev=2000000)

        except:
            log.warning("flipping_fit: the fitting was not succesful")
            popt = [0, 0, 2, 0]

        # sen fitting succesful
        if popt[2] != 2:
            eps = -1 / popt[2]
            amplitude_correction_factor = eps / (eps - 1)
            corrected_amplitude = amplitude_correction_factor * pi_pulse_amplitude
        # sen fitting not succesful = amplitude well adjusted
        else:
            amplitude_correction_factor = 1
            corrected_amplitude = amplitude_correction_factor * pi_pulse_amplitude

        corrected_amplitudes[qubit] = corrected_amplitude
        fitted_parameters[qubit] = popt
        amplitude_correction_factors[qubit] = amplitude_correction_factor

    return FlippingResults(
        corrected_amplitudes, amplitude_correction_factors, fitted_parameters
    )


def _plot(data: FlippngData, fit: FlippingResults, qubit):
    """Plotting function for Flipping."""

    figures = []
    fig = go.Figure()

    fitting_report = ""
    qubit_data = data.df[data.df["qubit"] == qubit]

    fig.add_trace(
        go.Scatter(
            x=qubit_data["flips"].pint.magnitude,
            y=qubit_data["MSR"].pint.to("uV").pint.magnitude,
            marker_color=get_color(0),
            opacity=1,
            name="Voltage",
            showlegend=True,
            legendgroup="Voltage",
        ),
    )

    # add fitting trace
    if len(data) > 0:
        flips_range = np.linspace(
            min(data.df["flips"]),
            max(data.df["flips"]),
            2 * len(data),
        )

        fig.add_trace(
            go.Scatter(
                x=flips_range,
                y=flipping_fit(
                    flips_range,
                    float(fit.fitted_parameters[qubit][0]),
                    float(fit.fitted_parameters[qubit][1]),
                    float(fit.fitted_parameters[qubit][2]),
                    float(fit.fitted_parameters[qubit][3]),
                ),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
                marker_color=get_color(1),
            ),
        )
        fitting_report = fitting_report + (
            f"q{qubit} | Amplitude_correction_factor: {fit.amplitude_factors[qubit]:.4f}<br>"
            + f"q{qubit} | Corrected_amplitude: {fit.amplitude[qubit][0]:.4f}<br><br>"
        )

    # last part
    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Flips (dimensionless)",
        yaxis_title="MSR (uV)",
    )

    figures.append(fig)

    return figures, fitting_report


flipping = Routine(_acquisition, _fit, _plot)
"""Flipping Routine  object."""



from qibocal.calibrations.characterization.allXY import *
from qibocal.calibrations.characterization.calibrate_qubit_states import *
from qibocal.calibrations.characterization.chevron import *
from qibocal.calibrations.characterization.cryoscope import *
from qibocal.calibrations.characterization.flipping import *
from qibocal.calibrations.characterization.qubit_spectroscopy import *
from qibocal.calibrations.characterization.rabi import *
from qibocal.calibrations.characterization.ramsey import *
from qibocal.calibrations.characterization.resonator_spectroscopy import *
from qibocal.calibrations.characterization.resonator_spectroscopy_sample import *
from qibocal.calibrations.characterization.ro_optimization import *
from qibocal.calibrations.characterization.spin_echo import *
from qibocal.calibrations.characterization.t1 import *



import datetime
import os
import re
import time

import pandas as pd
from dash import Dash, Input, Output, dcc, html

from qibocal import plots
from qibocal.calibrations.niGSC.basics.plot import plot_qq
from qibocal.data import DataUnits
from qibocal.web.server import server

DataUnits()  # dummy dataset call to suppress ``pint[V]`` error

app = Dash(
    server=server,
    suppress_callback_exceptions=True,
)

app.layout = html.Div(
    [
        html.Link(href="/static/styles.css", rel="stylesheet"),
        html.Div(
            id="refresh-area",
            className="refresh-area",
            children=[
                html.Div(
                    id="refresh-text",
                    className="refresh-text",
                    children=[html.P("Refresh rate:")],
                ),
                dcc.Dropdown(
                    id="interval-refresh",
                    className="interval-refresh",
                    placeholder="Select refresh rate",
                    options=[
                        {"label": "Auto", "value": 0},
                        {"label": "No refresh", "value": 3600},
                    ],
                    value=0,
                ),
                html.Div(
                    id="latest-timestamp",
                    className="latest-timestamp",
                ),
            ],
        ),
        html.Div(id="div-fitting", className="div-fitting"),
        html.Div(id="div-figures"),
        dcc.Location(id="url", refresh=False),
        dcc.Interval(
            id="interval",
            interval=160 * 1000,
            n_intervals=0,
            disabled=False,
        ),
    ]
)


@app.callback(
    Output(component_id="div-figures", component_property="children"),
    Output(component_id="latest-timestamp", component_property="children"),
    Output(component_id="interval", component_property="interval"),
    Output(component_id="div-fitting", component_property="children"),
    Input("interval", "n_intervals"),
    Input("url", "pathname"),
    Input("interval-refresh", "value"),
)
def get_graph(interval, url, value):
    st = time.time()

    figures = []

    if "data" not in url:
        url = f"/data{url}"

    method, folder, routine, qubit, format = url.split(os.sep)[2:]
    if qubit.isdigit():
        qubit = int(qubit)

    try:
        # data = DataUnits.load_data(folder, routine, format, "precision_sweep")
        # with open(f"{folder}/platform.yml", "r") as f:
        #     nqubits = yaml.safe_load(f)["nqubits"]
        # if len(data) > 2:
        #     params, fit = resonator_spectroscopy_fit(folder, format, nqubits)
        # else:
        #     params, fit = None, None
        # return getattr(plots.resonator_spectroscopy, method)(data, params, fit)

        # # FIXME: Temporarily hardcode the plotting method to test
        # # multiple routines with different names in one folder
        # # should be changed to:
        # # return getattr(getattr(plots, routine), method)(data)
        if hasattr(plots, method):
            figs, fitting_report = getattr(plots, method)(
                folder, routine, qubit, format
            )
        else:
            figs, fitting_report = plot_qq(folder, routine, qubit, format)
        et = time.time()

        if value == 0:
            refresh_rate = int(et - st) + 6
        else:
            refresh_rate = value

        for fig in figs:
            figures.append(dcc.Graph(figure=fig))

        table = ""
        if "No fitting data" not in fitting_report:
            fitting_params = re.split(r"<br>|:|\|", fitting_report)
            fitting_params = list(filter(lambda x: x.strip(), fitting_params))
            table_header = [
                html.Thead(
                    children=[
                        html.Tr(
                            children=[
                                html.Th(
                                    className="th_styles", children="qubit # / report #"
                                ),
                                html.Th(
                                    className="th_styles", children="Fitting Parameter"
                                ),
                                html.Th(className="th_styles", children="Value"),
                            ]
                        )
                    ]
                )
            ]
            table_rows = []

            for i in range(0, len(fitting_params), 3):
                table_rows.append(
                    html.Tr(
                        className="td_styles",
                        children=[
                            html.Td(fitting_params[i]),
                            html.Td(fitting_params[i + 1]),
                            html.Td(fitting_params[i + 2]),
                        ],
                    )
                )

            table = [
                html.Table(
                    className="fitting-table", children=table_header + table_rows
                )
            ]

        timestamp = datetime.datetime.now().strftime("%d/%m/%Y, %H:%M:%S")
        return (
            figures,
            [html.Span(f"Last update: {(timestamp)}")],
            refresh_rate * 1000,
            table,
        )
    except (FileNotFoundError, pd.errors.EmptyDataError):
        timestamp = datetime.datetime.now().strftime("%d/%m/%Y, %H:%M:%S")
        return (
            figures,
            [html.Span(f"Last updated: {timestamp}")],
            refresh_rate * 1000,
            table,
        )



import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from lmfit import Model
from plotly.subplots import make_subplots
from scipy.optimize import curve_fit
from scipy.signal import argrelextrema, lfilter, savgol_filter

from qibocal.data import Data, DataUnits
from qibocal.fitting.utils import cos, exp, flipping, lorenzian, rabi, ramsey
from qibocal.plots.utils import get_color, get_data_subfolders

MX_tag = "MX"
MY_tag = "MY"


def _get_flux_pulse_durations_and_amplitudes(data: DataUnits):
    flux_pulse_durations = data.get_values("flux_pulse_duration", "ns")[
        data.df["component"] == MX_tag
    ].to_numpy()
    flux_pulse_amplitudes = data.get_values("flux_pulse_amplitude", "dimensionless")[
        data.df["component"] == MX_tag
    ].to_numpy()
    amplitudes = flux_pulse_amplitudes[flux_pulse_durations == flux_pulse_durations[0]]
    durations = flux_pulse_durations[flux_pulse_amplitudes == flux_pulse_amplitudes[0]]
    return durations, amplitudes


def _get_values(data, component=None, duration=None, amplitude=None):
    if component:
        values = data.get_values("prob", "dimensionless")[
            data.df["component"] == component
        ]
    else:
        values = data.get_values("prob", "dimensionless")

    if duration:
        values = values[data.df["flux_pulse_duration"] == duration]
    if amplitude:
        values = values[data.df["flux_pulse_amplitude"] == amplitude]
    return _normalise(values.to_numpy())


def _normalise(values: np.array):
    values = ((values - np.min(values)) / (np.max(values) - np.min(values))) * 2 - 1

    # TODO: consider moving this normalisation to data acquisition (work with magnitudes instead of 0-1 probs)
    return values


def cryoscope_raw(folder, routine, qubit, format):
    figs = {}
    figs["raw"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, "data")
            # data.df = data.df[data.df["qubit"] == qubit]
            # extract qubit data and average over iterations
            data.df = (
                data.df[data.df["qubit"] == qubit]
                .drop(columns=["qubit", "iteration"])
                .groupby(
                    ["flux_pulse_duration", "flux_pulse_amplitude", "component"],
                    as_index=False,
                )
                .mean()
            )
        except:
            data = DataUnits(
                name=f"data",
                quantities={
                    "flux_pulse_duration": "ns",
                    "flux_pulse_amplitude": "dimensionless",
                    "prob": "dimensionless",
                },
                options=["component", "qubit", "iteration"],
            )

        durations = (
            data.df[data.df["component"] == MY_tag]["flux_pulse_duration"]
            .pint.to("ns")
            .pint.magnitude.unique()
        )
        amplitudes = (
            data.df[data.df["component"] == MY_tag]["flux_pulse_amplitude"]
            .pint.to("dimensionless")
            .pint.magnitude.unique()
        )

        for amp in amplitudes:
            figs["raw"].add_trace(
                go.Scatter(
                    x=durations,
                    y=_get_values(data, MX_tag, amplitude=amp),
                    name=f"q{qubit}/r{report_n}: <X> | A = {amp:.3f}",
                ),
                row=1,
                col=1,
            )

            figs["raw"].add_trace(
                go.Scatter(
                    x=durations,
                    y=_get_values(data, MY_tag, amplitude=amp),
                    name=f"q{qubit}/r{report_n}: <Y> | A = {amp:.3f}",
                ),
                row=1,
                col=1,
            )
    figs["raw"].update_layout(
        xaxis_title="Pulse duration (ns)",
        yaxis_title="Magnitude (dimensionless)",
        title=f"Raw data",
    )
    return [figs["raw"]]


def flux_pulse_timing(folder, routine, qubit, format):
    figs = {}
    figs["flux_pulse_timing"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    for subfolder in subfolders:
        try:
            data = DataUnits.load_data(folder, subfolder, routine, format, "data")
            data.df = (
                data.df[data.df["qubit"] == qubit]
                .drop(columns=["qubit", "iteration"])
                .groupby(
                    ["flux_pulse_amplitude", "flux_pulse_start"],
                    as_index=False,
                )
                .mean()
            )
        except:
            data = DataUnits(
                name=f"data",
                quantities={
                    "flux_pulse_amplitude": "dimensionless",
                    "flux_pulse_start": "ns",
                    "prob": "dimensionless",
                },
                options=["qubit", "iteration"],
            )

        amplitudes = (
            data.df["flux_pulse_amplitude"]
            .pint.to("dimensionless")
            .pint.magnitude.unique()
        )
        starts = data.df["flux_pulse_start"].pint.to("ns").pint.magnitude.unique()

    for amp in amplitudes:
        figs["flux_pulse_timing"].add_trace(
            go.Scatter(
                x=starts,
                y=_get_values(data, amplitude=amp),
                name=f"q{qubit}/r{report_n}: <X> | A = {amp:.3f}",
            ),
            row=1,
            col=1,
        )

    figs["flux_pulse_timing"].update_layout(
        xaxis_title="Flux start time (ns)",
        yaxis_title="Magnitude X component (dimensionless)",
    )
    return [figs["flux_pulse_timing"]]


def cryoscope_dephasing_heatmap(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    signal = []
    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)
        signal.append(re + 1j * im)
    signal = np.array(signal, dtype=np.complex128)
    phase = np.arctan2(signal.imag, signal.real)
    global_title = "Dephasing"
    title_x = "Flux Pulse Duration"
    title_y = "Amplitudes"
    figs["norm"].add_trace(
        go.Heatmap(
            x=durations,
            y=amplitudes,
            z=np.abs(signal),
            colorbar=dict(len=0.46, y=0.25),
        ),
        row=1,
        col=1,
    )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_fft_peak_fitting(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    signal = []  # TODO: replace lists with np arrays
    fft = []
    freqs = []

    curve_amps = []
    curve_detunings = []

    num_points = durations.shape[0]
    interval = (durations[1] - durations[0]) * 1e-9
    amp_freqs = np.fft.fftfreq(n=num_points, d=interval)
    mask = np.argsort(amp_freqs)
    amp_freqs = amp_freqs[mask]

    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        amp_signal = re + 1j * im
        signal.append(amp_signal)

        freqs.append(amp_freqs)

        amp_fft = np.fft.fft(amp_signal - np.mean(amp_signal))[mask]
        fft.append(amp_fft)

        figs["norm"].add_trace(
            go.Scatter(x=amp_freqs, y=np.abs(amp_fft), name=f"fft | A = {amp:.3f}"),
            row=1,
            col=1,
        )

        smoothed_y = smooth(np.absolute(amp_fft), window_len=21)
        percval = np.percentile(
            smoothed_y, 90
        )  # finds the top perc points in amplitude smoothed_y
        filtered_y = np.where(smoothed_y > percval, smoothed_y, percval)
        array_peaks = argrelextrema(filtered_y, np.greater)
        peaks_x = np.arange(num_points)[array_peaks]
        sort_mask = np.argsort(np.absolute(amp_fft)[array_peaks])[::-1]

        if len(peaks_x[sort_mask]) > 0 and amp_freqs[peaks_x[sort_mask][0]] <= 0:
            curve_amps.append(amp)
            curve_detunings.append(amp_freqs[peaks_x[sort_mask][0]])

        figs["norm"].add_trace(
            go.Scatter(x=amp_freqs, y=smoothed_y, name=f"smoothed_y | A = {amp:.3f}"),
            row=1,
            col=1,
        )

        figs["norm"].add_trace(
            go.Scatter(
                x=amp_freqs,
                y=filtered_y,
                name=f"filtered_y | A = {amp:.3f} | percentile = {percval:.3f}",
            ),
            row=1,
            col=1,
        )

    global_title = "FFT Peak Fitting"
    title_x = "Detunning (GHz)"
    title_y = "Spectral Density"

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_fft(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    signal = []  # TODO: replace lists with np arrays
    fft = []
    freqs = []

    curve_amps = []
    curve_detunings = []

    num_points = durations.shape[0]
    interval = (durations[1] - durations[0]) * 1e-9
    amp_freqs = np.fft.fftfreq(n=num_points, d=interval)
    mask = np.argsort(amp_freqs)
    amp_freqs = amp_freqs[mask]

    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        amp_signal = re + 1j * im
        signal.append(amp_signal)

        freqs.append(amp_freqs)

        amp_fft = np.fft.fft(amp_signal - np.mean(amp_signal))[mask]
        fft.append(amp_fft)

        smoothed_y = smooth(np.absolute(amp_fft), window_len=21)
        percval = np.percentile(
            smoothed_y, 90
        )  # finds the top perc points in amplitude smoothed_y
        filtered_y = np.where(smoothed_y > percval, smoothed_y, percval)
        array_peaks = argrelextrema(filtered_y, np.greater)
        peaks_x = np.arange(num_points)[array_peaks]
        sort_mask = np.argsort(np.absolute(amp_fft)[array_peaks])[::-1]

        if len(peaks_x[sort_mask]) > 0 and amp_freqs[peaks_x[sort_mask][0]] <= 0:
            curve_amps.append(amp)
            curve_detunings.append(amp_freqs[peaks_x[sort_mask][0]])

    # print(curve_amps,curve_detunings)
    pfit = np.polyfit(curve_amps, curve_detunings, 2)
    # DEBUG:
    _amp = -0.109
    print(f"Curve fit {pfit} _ {pfit[0]*_amp**2 + pfit[1]*_amp + pfit[2]}")

    curve_amps = np.array(curve_amps)

    global_title = "FFT with curve fit"
    title_x = "Detunning (GHz)"
    title_y = "Amplitudes"
    figs["norm"].add_trace(
        go.Heatmap(
            x=amp_freqs,
            y=amplitudes,
            z=np.abs(fft),
        ),
        row=1,
        col=1,
    )
    figs["norm"].add_trace(
        go.Scatter(
            x=curve_detunings,
            y=curve_amps,
            name=f"Data points",
            mode="markers",
            marker=dict(size=5),
        ),
        row=1,
        col=1,
    )
    figs["norm"].add_trace(
        go.Scatter(
            x=pfit[0] * curve_amps**2 + pfit[1] * curve_amps + pfit[2],
            y=curve_amps,
            name=f"Curve fit",
        ),
        row=1,
        col=1,
    )
    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_phase(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        phase = np.arctan2(im, re)

        # Phase vs. Time
        global_title = "Phase vs. Time"
        title_x = "Flux Pulse duration"
        title_y = "Phase"
        figs["norm"].add_trace(
            go.Scatter(
                x=durations,
                y=phase,
                name=f"Phase | A = {amp:.3f}",
                mode="lines+markers",
                marker=dict(size=5),
            ),
            row=1,
            col=1,
        )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_phase_heatmap(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    signal = []
    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)
        signal.append(re + 1j * im)
    signal = np.array(signal, dtype=np.complex128)
    phase = np.arctan2(signal.imag, signal.real)

    global_title = "Phase vs. Time"
    title_x = "Flux Pulse duration"
    title_y = "Amplitude"
    figs["norm"].add_trace(
        go.Heatmap(
            x=durations,
            y=amplitudes,
            z=phase,
        ),
        row=1,
        col=1,
    )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_phase_unwrapped(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        phase = np.arctan2(im, re)
        phase_unwrapped = np.unwrap(phase)

        # Phase vs. Time
        global_title = "Phase Unwrapped vs. Time"
        title_x = "Flux Pulse duration"
        title_y = "Phase"
        figs["norm"].add_trace(
            go.Scatter(x=durations, y=phase_unwrapped, name=f"Phase | A = {amp:.3f}"),
            row=1,
            col=1,
        )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_phase_unwrapped_heatmap(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    signal = []
    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)
        signal.append(re + 1j * im)
    signal = np.array(signal, dtype=np.complex128)
    phase = np.arctan2(signal.imag, signal.real)
    phase_unwrapped = np.unwrap(phase, axis=1)

    global_title = "Phase unwrapped along duration vs. Time"
    title_x = "Flux Pulse duration"
    title_y = "Phase"
    figs["norm"].add_trace(
        go.Heatmap(
            x=durations,
            y=amplitudes,
            z=phase_unwrapped,
        ),
        row=1,
        col=1,
    )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_phase_amplitude_unwrapped_heatmap(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    # start the unroll desde amp = 0 (no phase shift)
    # unroll in both directions

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    signal = []
    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)
        signal.append(re + 1j * im)
    signal = np.array(signal, dtype=np.complex128)
    phase = np.arctan2(signal.imag, signal.real)
    phase_unwrapped = np.unwrap(phase, axis=1)

    global_title = "Phase unwrapped along amplitude vs. Time"
    title_x = "Flux Pulse duration"
    title_y = "Phase"
    phase = np.arctan2(signal.imag, signal.real)
    phase_unwrapped = np.unwrap(phase, axis=0)
    figs["norm"].add_trace(
        go.Heatmap(
            x=durations,
            y=amplitudes,
            z=phase_unwrapped,
        ),
        row=1,
        col=1,
    )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_detuning_time(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    sampling_rate = 1 / (durations[1] - durations[0])
    derivative_window_length = 7 / sampling_rate
    derivative_window_size = max(3, int(derivative_window_length * sampling_rate))
    derivative_window_size += (derivative_window_size + 1) % 2
    derivative_window_size = 9
    derivative_order = 2
    nyquist_order = 0

    # signal = []
    # detuning_mean = []
    # signal_fft = []

    for amp in amplitudes:  # unwrap along duration
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        # Unwrap phase
        phase = np.arctan2(im, re)
        phase_unwrapped = np.unwrap(phase)

        # -----------------      Unwrap along durations     ---------------------------
        # phase_unwrapped_matrix = np.array(amplitudes).reshape(len(amplitudes),1)
        # for dur in durations: # unwrap along amplitudes
        #     re = _get_values(data, MX_tag, duration=dur)
        #     im = _get_values(data, MY_tag, duration=dur)

        #     # Unwrap phase
        #     phase = np.arctan2(im, re)
        #     phase_unwrapped = np.unwrap(phase)

        #     phase_unwrapped_matrix = np.append(phase_unwrapped_matrix, phase_unwrapped, axis=1)

        # for n in range(1, len(amplitudes)):
        #     phase_unwrapped = phase_unwrapped_matrix[n]
        # -----------------------------------------------------------------------------

        # use a savitzky golay filter: it takes sliding window of length
        # `window_length`, fits a polynomial, returns derivative at
        # middle point

        # Derivate
        detuning = (
            savgol_filter(
                phase_unwrapped / (2 * np.pi),
                window_length=derivative_window_size,
                polyorder=derivative_order,
                deriv=1,
            )
            * sampling_rate
            * 1e9
        )

        # nyquist_order = 0
        # detuning = get_real_detuning(detuning, demod_freq, sampling_rate, nyquist_order)

        # Alternative derivation method
        # dt = np.diff(duration_unique) * 1e-9
        # dphi_dt_unwrap = np.diff(phase_unwrapped) / dt
        # detuning = dphi_dt_unwrap / (2 * np.pi)
        # detuning = smooth(detuning, window_len=21)

        # Detunning vs. Time
        global_title = "Detuning vs. Time"
        title_x = "Flux Pulse duration"
        title_y = "Detunning"
        figs["norm"].add_trace(
            go.Scatter(x=durations, y=detuning, name=f"A = {amp:.3f}"),
            row=1,
            col=1,
        )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_distorted_amplitude_time(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    signal = []  # TODO: replace lists with np arrays
    fft = []
    freqs = []

    curve_amps = []
    curve_detunings = []

    num_points = durations.shape[0]
    interval = (durations[1] - durations[0]) * 1e-9
    amp_freqs = np.fft.fftfreq(n=num_points, d=interval)
    mask = np.argsort(amp_freqs)
    amp_freqs = amp_freqs[mask]

    sampling_rate = 1 / (durations[1] - durations[0])
    derivative_window_length = 7 / sampling_rate
    derivative_window_size = max(3, int(derivative_window_length * sampling_rate))
    derivative_window_size += (derivative_window_size + 1) % 2
    derivative_window_size = 9
    derivative_order = 3
    nyquist_order = 0

    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        amp_signal = re + 1j * im
        signal.append(amp_signal)

        freqs.append(amp_freqs)

        amp_fft = np.fft.fft(amp_signal - np.mean(amp_signal))[mask]
        fft.append(amp_fft)

        smoothed_y = smooth(np.absolute(amp_fft), window_len=21)
        percval = np.percentile(
            smoothed_y, 90
        )  # finds the top perc points in amplitude smoothed_y
        filtered_y = np.where(smoothed_y > percval, smoothed_y, percval)
        array_peaks = argrelextrema(filtered_y, np.greater)
        peaks_x = np.arange(num_points)[array_peaks]
        sort_mask = np.argsort(np.absolute(amp_fft)[array_peaks])[::-1]

        if len(peaks_x[sort_mask]) > 0 and amp_freqs[peaks_x[sort_mask][0]] <= 0:
            curve_amps.append(amp)
            curve_detunings.append(amp_freqs[peaks_x[sort_mask][0]])

    pfit = np.polyfit(curve_amps, curve_detunings, 2)
    curve_amps = np.array(curve_amps)

    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        # Unwrap phase
        phase = np.arctan2(im, re)
        phase_unwrapped = np.unwrap(phase)

        # -----------------      Unwrap along durations     ---------------------------
        # phase_unwrapped_matrix = np.array(amplitudes).reshape(len(amplitudes),1)
        # for dur in durations: # unwrap along amplitudes
        #     re = _get_values(data, MX_tag, duration=dur)
        #     im = _get_values(data, MY_tag, duration=dur)

        #     # Unwrap phase
        #     phase = np.arctan2(im, re)
        #     phase_unwrapped = np.unwrap(phase)

        #     phase_unwrapped_matrix = np.append(phase_unwrapped_matrix, phase_unwrapped, axis=1)

        # for n in range(1, len(amplitudes)):
        #     phase_unwrapped = phase_unwrapped_matrix[n]
        # -----------------------------------------------------------------------------

        # Derivate
        detuning = (
            savgol_filter(
                phase_unwrapped / (2 * np.pi),
                window_length=derivative_window_size,
                polyorder=derivative_order,
                deriv=1,
            )
            * sampling_rate
            * 1e9
        )

        clipped_detuning = np.clip(
            detuning,
            a_max=-1 / 4 * (pfit[1] ** 2) / pfit[0] + pfit[2],
            a_min=np.min(detuning),
        )
        plt.plot(detuning)
        plt.plot(clipped_detuning)
        plt.savefig("/nfs/users/alvaro.orgaz/clipped.png")
        plt.close()

        distorted_pulse = (
            -pfit[1]
            - np.sqrt(pfit[1] ** 2 - (4 * pfit[0] * (pfit[2] - clipped_detuning)))
        ) / (2 * pfit[0])
        # distorted_pulse = (-pfit[1] + np.sqrt((pfit[1]**2 - (4* pfit[0] * (pfit[2] - clipped_detuning)))))/(2*pfit[0])

        global_title = "Distorted Pulse Amplitude vs. Time"
        title_x = "Flux Pulse duration"
        title_y = "Amplitude"
        figs["norm"].add_trace(
            go.Scatter(
                x=durations,
                y=distorted_pulse,
                name=f"A = {amp:.3f}",
                mode="lines+markers",
                marker=dict(size=5),
            ),
            row=1,
            col=1,
        )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


def cryoscope_reconstructed_amplitude_time(folder, routine, qubit, format):
    data = DataUnits.load_data(folder, "data", routine, format, "data")
    durations, amplitudes = _get_flux_pulse_durations_and_amplitudes(data)

    figs = {}
    figs["norm"] = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.2,
    )

    signal = []  # TODO: replace lists with np arrays
    fft = []
    freqs = []

    curve_amps = []
    curve_detunings = []

    num_points = durations.shape[0]
    interval = (durations[1] - durations[0]) * 1e-9
    amp_freqs = np.fft.fftfreq(n=num_points, d=interval)
    mask = np.argsort(amp_freqs)
    amp_freqs = amp_freqs[mask]

    sampling_rate = 1 / (durations[1] - durations[0])
    derivative_window_length = 7 / sampling_rate
    derivative_window_size = max(3, int(derivative_window_length * sampling_rate))
    derivative_window_size += (derivative_window_size + 1) % 2
    derivative_window_size = 15
    derivative_order = 3
    nyquist_order = 0

    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        amp_signal = re + 1j * im
        signal.append(amp_signal)

        freqs.append(amp_freqs)

        amp_fft = np.fft.fft(amp_signal - np.mean(amp_signal))[mask]
        fft.append(amp_fft)

        smoothed_y = smooth(np.absolute(amp_fft), window_len=21)
        percval = np.percentile(
            smoothed_y, 90
        )  # finds the top perc points in amplitude smoothed_y
        filtered_y = np.where(smoothed_y > percval, smoothed_y, percval)
        array_peaks = argrelextrema(filtered_y, np.greater)
        peaks_x = np.arange(num_points)[array_peaks]
        sort_mask = np.argsort(np.absolute(amp_fft)[array_peaks])[::-1]

        if len(peaks_x[sort_mask]) > 0 and amp_freqs[peaks_x[sort_mask][0]] <= 0:
            curve_amps.append(amp)
            curve_detunings.append(amp_freqs[peaks_x[sort_mask][0]])

    pfit = np.polyfit(curve_amps, curve_detunings, 2)
    curve_amps = np.array(curve_amps)
    # x=pfit[0]*curve_amps**2 + pfit[1]*curve_amps + pfit[2]

    for amp in amplitudes:
        re = _get_values(data, MX_tag, amplitude=amp)
        im = _get_values(data, MY_tag, amplitude=amp)

        # Unwrap phase
        phase = np.arctan2(im, re)
        phase_unwrapped = np.unwrap(phase)

        # -----------------      Unwrap along durations     ---------------------------
        # phase_unwrapped_matrix = np.array(amplitudes).reshape(len(amplitudes),1)
        # for dur in durations: # unwrap along amplitudes
        #     re = _get_values(data, MX_tag, duration=dur)
        #     im = _get_values(data, MY_tag, duration=dur)

        #     # Unwrap phase
        #     phase = np.arctan2(im, re)
        #     phase_unwrapped = np.unwrap(phase)

        #     phase_unwrapped_matrix = np.append(phase_unwrapped_matrix, phase_unwrapped, axis=1)

        # for n in range(1, len(amplitudes)):
        #     phase_unwrapped = phase_unwrapped_matrix[n]
        # -----------------------------------------------------------------------------

        # Derivate
        detuning = (
            savgol_filter(
                phase_unwrapped / (2 * np.pi),
                window_length=derivative_window_size,
                polyorder=derivative_order,
                deriv=1,
            )
            * sampling_rate
            * 1e9
        )  # TODO: fix units e28!!??
        print(f"pfit: {pfit}")

        clipped_detuning = np.clip(
            detuning,
            a_max=-1 / 4 * (pfit[1] ** 2) / pfit[0] + pfit[2],
            a_min=np.min(detuning),
        )
        plt.plot(detuning)
        plt.plot(clipped_detuning)
        plt.savefig("/nfs/users/alvaro.orgaz/clipped.png")
        plt.close()

        distorted_pulse = (
            -pfit[1]
            - np.sqrt(pfit[1] ** 2 - (4 * pfit[0] * (pfit[2] - clipped_detuning)))
        ) / (2 * pfit[0])
        # distorted_pulse = (-pfit[1] + np.sqrt((pfit[1]**2 - (4* pfit[0] * (pfit[2] - clipped_detuning)))))/(2*pfit[0])

        global_title = "Reconstructed Pulse Amplitude vs. Time"
        title_x = "Flux Pulse duration"
        title_y = "Amplitude"
        figs["norm"].add_trace(
            go.Scatter(
                x=durations,
                y=distorted_pulse,
                name=f"Distorted A = {amp:.3f}",
                mode="lines+markers",
                marker=dict(size=5),
            ),
            row=1,
            col=1,
        )

        # fit over amplitudes vs time
        ideal_pulse = np.ones(len(durations))
        start_fit = 40 - 4
        end_fit = 165 - 4  # len(durations)
        amp_guess = -0.78
        tau_guess = 37
        pguess = [amp_guess, tau_guess]
        b_fit, a_fit, popt = cryoscope_fit_exp(
            durations=durations[start_fit:end_fit],
            ideal_pulse=ideal_pulse[start_fit:end_fit],
            distorted_pulse=distorted_pulse[start_fit:end_fit],
            pguess=pguess,
        )  # FIXME
        # print(b_fit, a_fit)
        reconstructed_pulse = cryoscope_step_response_exp(
            t=durations[start_fit:end_fit], p=popt, pulse=ideal_pulse[start_fit:end_fit]
        )

        # simulated_step_extrap = cryoscope_step_response_exp(t=durations,
        #                                             p=popt,
        #                                             awg_reconstructed_pulse=ideal_pulse)

        # initial_guess = cryoscope_step_response_exp(t=durations,
        #                                             p=[0.78,37],
        #                                             awg_reconstructed_pulse=ideal_pulse)

        figs["norm"].add_trace(
            go.Scatter(
                x=durations[start_fit:end_fit],
                y=ideal_pulse,
                name=f"Ideal A = {amp:.3f}",
                mode="lines+markers",
                marker=dict(size=5),  # , color="lightcoral"),
            ),
            row=1,
            col=1,
        )

        figs["norm"].add_trace(
            go.Scatter(
                x=durations[start_fit:end_fit],
                y=reconstructed_pulse,
                name=f"Reconstructed A = {amp:.3f}",
                mode="lines+markers",
                marker=dict(size=5),  # , color="lightcoral"),
            ),
            row=1,
            col=1,
        )

    figs["norm"].update_layout(
        xaxis_title=title_x,
        yaxis_title=title_y,
        title=f"{global_title}",
    )
    return [figs["norm"]]


# Helper functions


def smooth(x, window_len=11, window="hanning"):
    """smooth the data using a window with requested size.
    This method is based on the convolution of a scaled window with the
        signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are
        minimized
    in the begining and end part of the output signal.
    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an
            odd integer
        window: the type of window from 'flat', 'hanning', 'hamming',
            'bartlett', 'blackman'
        flat window will produce a moving average smoothing.
    output:
        the smoothed signal
    example:
    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    see also:
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman,
    numpy.convolve
    scipy.signal.lfilter
    """
    if int(window_len) & 0x1 == 0:
        window_len += 1

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if not window in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
        raise ValueError(
            "Window is one of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"
        )

    s = np.r_[x[window_len - 1 : 0 : -1], x, x[-1:-window_len:-1]]

    if window == "flat":  # moving average
        w = np.ones(window_len, "d")

    else:
        w = eval("np." + window + "(window_len)")
    y = np.convolve(w / w.sum(), s, mode="valid")

    # Cut edges of y since a mirror image is used
    edge = (window_len - 1) / 2
    edge = int(edge)
    return y[edge:-edge]


def coefs_from_exp(amp, tau):
    alpha = 1 - np.exp(-1 / (tau * (1 + amp)))
    # the coefficient k (and the filter transfer function)
    # depend on the sign of amp

    if amp >= 0.0:
        k = amp / (1 + amp - alpha)
        # compensating overshoot by a low-pass filter
        # correction filter y[n] = (1-k)*x[n] + k*u[n] = x[n] + k*(u[n]-x[n])
        # where u[n] = u[n-1] + alpha*(x[n] - u[n-1])
    else:
        k = amp / (1 + amp) / (1 - alpha)
        # compensating low-pass by an overshoot
        # correction filter y[n] = (1+k)*x[n] - k*u[n] = x[n] - k*(u[n]-x[n])
        # where u[n] = u[n-1] + alpha*(x[n] - u[n-1])

    b = [(1 - k + k * alpha), -(1 - k) * (1 - alpha)]

    # while the denominator stays the same
    a = [1, -(1 - alpha)]
    return b, a


def cryoscope_step_response_exp(t, p, pulse):
    # This function returns the convolution of a filer h(t,params=p)
    #   applied to meassured distorted pulse -> should return rectangular
    #   applied to a rectangular pulse -> returns a predistorted pulse
    # It is done via lfilter function, and using an exponential model for h(t,p)

    # https://arxiv.org/pdf/1907.04818.pdf (page 11 - filter formula S22)
    # p = [A, tau_iir]
    # p = [b0 = 1−k +k ·α, b1 = −(1−k)·(1−α),a0 = 1 and a1 = −(1−α)]
    # p = [b0, b1, a0, a1]
    from scipy.signal import lfilter

    amp = p[0]  # ration between the shooting point and the settling amplitude
    tau_iir = p[1]
    b_fit, a_fit = coefs_from_exp(amp, tau_iir)
    return lfilter(b_fit, a_fit, pulse)


def cryoscope_fit_exp(durations, ideal_pulse, distorted_pulse, pguess):
    # print('calling curve_fit')
    # print(pguess)
    amp_guess = pguess[0]
    tau_guess = pguess[1]

    def wrapper(t, amp, tau, gain):
        return gain * cryoscope_step_response_exp(t, [amp, tau], distorted_pulse)

    fit_mod = Model(wrapper)
    fit_mod.set_param_hint("amp", value=amp_guess, min=-2, max=2, vary=True)
    fit_mod.set_param_hint("tau", value=tau_guess, min=0, max=1000, vary=True)
    fit_mod.set_param_hint("gain", value=1, min=0, max=2, vary=True)
    params = fit_mod.make_params()
    fit_result = fit_mod.fit(data=ideal_pulse, t=durations, params=params)
    print(fit_result.fit_report())
    # popt, pcov = curve_fit(, x_data, y_data, p0=pguess)
    print(fit_result.best_values)
    fit_result.plot_fit()
    amp = fit_result.best_values[
        "amp"
    ]  # ration between the shooting point and the settling amplitude
    tau_iir = fit_result.best_values["tau"]
    b_fit, a_fit = coefs_from_exp(amp, tau_iir)
    return b_fit, a_fit, [amp, tau_iir]


# def cryoscope_fit(x_data, y_data, awg_reconstructed_pulse):
#     pguess = [
#         1, #b0
#         0, #b1
#         1, #a0
#         0  #a1
#     ]
#     popt, pcov = curve_fit(lambda t,*p:cryoscope_step_response(t, p, awg_reconstructed_pulse), x_data, y_data, p0=pguess)
#     b_fit = popt[[0,1]]
#     a_fit = popt[[2,3]]
#     return b_fit, a_fit

# def cryoscope_step_response(t, p, awg_reconstructed_pulse):

#     # https://arxiv.org/pdf/1907.04818.pdf (page 11 - filter formula S22)
#     # p = [b0 = 1−k +k ·α, b1 = −(1−k)·(1−α),a0 = 1 and a1 = −(1−α)]
#     # p = [b0, b1, a0, a1]
#     from scipy.signal import lfilter
#     b_fit = p[:2]
#     a_fit = p[2:]
#     return lfilter(b_fit, a_fit, awg_reconstructed_pulse)

# def cryoscope_step_response_exp(t, p, awg_reconstructed_pulse):

#     # https://arxiv.org/pdf/1907.04818.pdf (page 11 - filter formula S22)
#     # p = [A, tau_iir]
#     # p = [b0 = 1−k +k ·α, b1 = −(1−k)·(1−α),a0 = 1 and a1 = −(1−α)]
#     # p = [b0, b1, a0, a1]
#     from scipy.signal import lfilter
#     amp = p[0] # ration between the shooting point and the settling amplitude
#     tau_iir = p[1]
#     alpha = 1-np.exp((1+amp)/tau_iir)
#     if amp<0:
#         k = amp/((1+amp)*(1-alpha))
#     else:
#         k = 1/(1+amp-alpha)
#     b_fit = [1-k+k*alpha, -(1-k)*(1-alpha)]
#     a_fit = [1,-(1-alpha)]
#     return lfilter(b_fit, a_fit, awg_reconstructed_pulse)

# def cryoscope_fit_exp(x_data, y_data, awg_reconstructed_pulse):
#     amp_guess = -0.78
#     tau_guess = 37
#     pguess = [amp_guess, tau_guess]
#     popt, pcov = curve_fit(lambda t,*p:cryoscope_step_response_exp(t, p, awg_reconstructed_pulse), x_data, y_data, p0=pguess)
#     amp = popt[0] # ration between the shooting point and the settling amplitude
#     tau_iir = popt[1]
#     alpha = 1-np.exp((1+amp)/tau_iir)
#     if amp<0:
#         k = amp/((1+amp)*(1-alpha))
#     else:
#         k = 1/(1+amp-alpha)
#     b_fit = [1-k+k*alpha, -(1-k)*(1-alpha)]
#     a_fit = [1,-(1-alpha)]
#     return b_fit, a_fit, popt

# ?
# def cryoscope_detuning_amplitude(folder, routine, qubit, format):
#     data = DataUnits.load_data(folder, "data", routine, format, "data")
#     import numpy as np
#     import scipy.signal as ss

#     MX_tag = "MX"
#     MY_tag = "MY"

#     amplitude = data.get_values("flux_pulse_amplitude", "dimensionless")
#     duration = data.get_values("flux_pulse_duration", "ns")
#     flux_pulse_duration = duration[data.df["component"] == MY_tag].to_numpy()
#     flux_pulse_amplitude = amplitude[data.df["component"] == MY_tag].to_numpy()
#     amplitude_unique = flux_pulse_amplitude[
#         flux_pulse_duration == flux_pulse_duration[0]
#     ]
#     duration_unique = flux_pulse_duration[
#         flux_pulse_amplitude == flux_pulse_amplitude[0]
#     ]

#     # Making figure
#     figs = {}
#     figs["norm"] = make_subplots(
#         rows=1,
#         cols=1,
#         horizontal_spacing=0.1,
#         vertical_spacing=0.2,
#     )

#     mean_detuning = []

#     for amp in amplitude_unique:
#         x_data = data.get_values("prob", "dimensionless")[data.df["flux_pulse_amplitude"] == amp][data.df["component"] == MX_tag].to_numpy()
#         y_data = data.get_values("prob", "dimensionless")[data.df["flux_pulse_amplitude"] == amp][data.df["component"] == MY_tag].to_numpy()

#         complex_data = x_data + 1j * y_data
#         #print (len(complex_data))
#         ws = 10 # Needs to be adjusted for data analysis
#         norm_data = normalize_sincos(complex_data, window_size=ws)

#         sampling_rate = 1 / (duration_unique[1] - duration_unique[0])
#         derivative_window_length = 7 / sampling_rate
#         derivative_window_size = max(3, int(derivative_window_length * sampling_rate))
#         derivative_window_size += (derivative_window_size + 1) % 2
#         derivative_order = 2
#         nyquist_order=0

#         freq_guess, phase_guess, offset_guess, amp_guess = fft_based_freq_guess_complex(norm_data)
#         demod_freq = - \
#                 freq_guess * sampling_rate

#         #y = amp*exp(2pi i f t + ph) + off.
#         #demod_data = amp_guess * np.exp(2 * np.pi * 1j * demod_freq * duration_unique + phase_guess) + offset_guess
#         demod_data = np.exp(2 * np.pi * 1j * duration_unique * demod_freq) * norm_data
#         # Di Carlo has smooth here!!!


#         # Unwrap phase
#         phase = np.arctan2(demod_data.imag, demod_data.real)
#         # phase_unwrapped = np.unwrap(phase)
#         phase_unwrapped = np.unwrap(np.angle(demod_data))

#         # use a savitzky golay filter: it take sliding window of length
#         # `window_length`, fits a polynomial, returns derivative at
#         # middle point
#         # phase = phase_unwrapped_data

#         # Di Carlo method
#         detuning = savgol_filter(
#             phase_unwrapped / (2 * np.pi),
#             window_length = derivative_window_size,
#             polyorder = derivative_order,
#             deriv=1) * sampling_rate * 1e9

#         nyquist_order = 0
#         detuning = get_real_detuning(detuning, demod_freq, sampling_rate, nyquist_order)

#         # Maxime method
#         # dt = np.diff(duration_unique) * 1e-9
#         # dphi_dt_unwrap = np.abs(np.diff(phase_unwrapped) / dt)
#         # detuning = dphi_dt_unwrap / (2 * np.pi)


#         mean_detuning.append(np.mean(detuning))

#     # Mean detunning vs. amplitude
#     global_title = "Mean Detuning vs. Amplitude"
#     title_x = "Amplitude (dimensionless)"
#     title_y = "Detunning mean (Hz)"
#     figs["norm"].add_trace(
#         go.Scatter(
#             x=amplitude_unique,
#             y=mean_detuning,
#             name=f"Detuning"
#         ),
#         row=1,
#         col=1,
#     )

#     figs["norm"].update_layout(
#         xaxis_title=title_x,
#         yaxis_title=title_y,
#         title=f"{global_title}",
#     )
#     return [figs["norm"]]


# def peak_finder_v2(x, y, perc=90, window_len=11):
#     """
#     Peak finder based on argrelextrema function from scipy
#     only finds maximums, this can be changed to minimum by using -y instead of y
#     """
#     smoothed_y = smooth(y, window_len=window_len)
#     percval = np.percentile(smoothed_y, perc) # finds the top perc points in amplitude smoothed_y
#     filtered_y = np.where(smoothed_y > percval, smoothed_y, percval)
#     array_peaks = argrelextrema(filtered_y, np.greater)
#     peaks_x = x[array_peaks]
#     sort_mask = np.argsort(y[array_peaks])[::-1]
#     return peaks_x[sort_mask]


# def normalize_sincos(
#         data,
#         window_size_frac=500,
#         window_size=None,
#         do_envelope=True):

#     if window_size is None:
#         window_size = len(data) // window_size_frac

#         # window size for savgol filter must be odd
#         window_size -= (window_size + 1) % 2

#     mean_data_r = savgol_filter(data.real, window_size, 0, 0)
#     mean_data_i = savgol_filter(data.imag, window_size, 0, 0)

#     mean_data = mean_data_r + 1j * mean_data_i

#     if do_envelope:
#         envelope = np.sqrt(
#             savgol_filter(
#                 (np.abs(
#                     data -
#                     mean_data))**2,
#                 window_size,
#                 0,
#                 0))
#     else:
#         envelope = 1
#     norm_data = ((data - mean_data) / envelope)
#     return norm_data

# def fft_based_freq_guess_complex(y):
#     """
#     guess the shape of a sinusoidal complex signal y (in multiples of
#         sampling rate), by selecting the peak in the fft.
#     return guess (f, ph, off, amp) for the model
#         y = amp*exp(2pi i f t + ph) + off.
#     """
#     fft = np.fft.fft(y)[1:len(y)]
#     freq_guess_idx = np.argmax(np.abs(fft))
#     if freq_guess_idx >= len(y) // 2:
#         freq_guess_idx -= len(y)

#     freq_guess = 1 / len(y) * (freq_guess_idx + 1)
#     phase_guess = np.angle(fft[freq_guess_idx]) + np.pi / 2
#     amp_guess = np.absolute(fft[freq_guess_idx]) / len(y)
#     offset_guess = np.mean(y)

#     return freq_guess, phase_guess, offset_guess, amp_guess

# def get_real_detuning(detuning, demod_freq, sampling_rate, nyquist_order=0):
#     real_detuning = detuning - demod_freq + sampling_rate * nyquist_order
#     return real_detuning

# def normalize_data(x, y, amplitude_unique, duration_unique, flux_pulse_amplitude, flux_pulse_duration, amplitude):
#     x_norm = np.ones((len(amplitude_unique), len(duration_unique))) * np.nan
#     y_norm = np.ones((len(amplitude_unique), len(duration_unique))) * np.nan

#     for i, amp in enumerate(amplitude_unique):
#         idx = np.where(flux_pulse_amplitude == amp)
#         if amp == flux_pulse_amplitude[-1]:
#             n = int(np.where(flux_pulse_duration[-1] == duration_unique)[0][0]) + 1
#         else:
#             n = len(duration_unique)

#         x_norm[i, :n] = x[idx]
#         x_norm[i, :n] = x_norm[i, :n] - 0.5
#         x_norm[i, :n] = x_norm[i, :n] / max(abs(x_norm[i, :n]))
#         y_norm[i, :n] = y[idx]
#         y_norm[i, :n] = y_norm[i, :n] - 0.5
#         y_norm[i, :n] = y_norm[i, :n] / max(abs(y_norm[i, :n]))

#     return x_norm, y_norm

# def normalize_sincos(x, y, window_size_frac=61, window_size=None, do_envelope=True):

#     if window_size is None:
#         window_size = len(x) // window_size_frac

#         # window size for savgol filter must be odd
#         window_size -= (window_size + 1) % 2

#     x = savgol_filter(x, window_size, 0, 0)  ## why poly order 0
#     y = savgol_filter(y, window_size, 0, 0)

#     # if do_envelope:
#     #     envelope = np.sqrt(savgol_filter((np.abs(data - mean_data))**2, window_size, 0, 0))
#     # else:
#     #     envelope = 1

#     # norm_data = ((data - mean_data) / envelope)

#     return x, y

# def cryoscope_demod_circle(folder, routine, qubit, format):
#     data = DataUnits.load_data(folder, "data", routine, format, "data")
#     import numpy as np
#     import scipy.signal as ss

#     MX_tag = "MX"
#     MY_tag = "MY"

#     amplitude = data.get_values("flux_pulse_amplitude", "dimensionless")
#     duration = data.get_values("flux_pulse_duration", "ns")
#     flux_pulse_duration = duration[data.df["component"] == MY_tag].to_numpy()
#     flux_pulse_amplitude = amplitude[data.df["component"] == MY_tag].to_numpy()
#     amplitude_unique = flux_pulse_amplitude[
#         flux_pulse_duration == flux_pulse_duration[0]
#     ]
#     duration_unique = flux_pulse_duration[
#         flux_pulse_amplitude == flux_pulse_amplitude[0]
#     ]

#     # Making figure
#     figs = {}
#     figs["norm"] = make_subplots(
#         rows=1,
#         cols=1,
#         horizontal_spacing=0.1,
#         vertical_spacing=0.2,
#     )

#     for amp in amplitude_unique:
#         x_data = data.get_values("prob", "dimensionless")[data.df["flux_pulse_amplitude"] == amp][data.df["component"] == MX_tag].to_numpy()
#         y_data = data.get_values("prob", "dimensionless")[data.df["flux_pulse_amplitude"] == amp][data.df["component"] == MY_tag].to_numpy()

#         complex_data = x_data + 1j * y_data
#         #print (len(complex_data))
#         ws = 10 # Needs to be adjusted for data analysis
#         norm_data = normalize_sincos(complex_data, window_size=ws)

#         sampling_rate = 1 / (duration_unique[1] - duration_unique[0])
#         derivative_window_length = 7 / sampling_rate
#         derivative_window_size = max(3, int(derivative_window_length * sampling_rate))
#         derivative_window_size += (derivative_window_size + 1) % 2
#         derivative_order = 2
#         nyquist_order=0

#         freq_guess, phase_guess, offset_guess, amp_guess = fft_based_freq_guess_complex(norm_data)
#         demod_freq = - \
#                 freq_guess * sampling_rate

#         #y = amp*exp(2pi i f t + ph) + off.
#         #demod_data = amp_guess * np.exp(2 * np.pi * 1j * demod_freq * duration_unique + phase_guess) + offset_guess
#         demod_data = np.exp(2 * np.pi * 1j * duration_unique * demod_freq) * norm_data
#         # Di Carlo has smooth here!!!

#         # demod_data circle
#         global_title = "FFT demod data circle"
#         title_x = "<X>"
#         title_y = "<Y>"
#         figs["norm"].add_trace(
#             go.Scatter(
#                 x=demod_data.real,
#                 y=demod_data.imag,
#                 name=f"A = {amp:.3f}"
#             ),
#             row=1,
#             col=1,
#         )

#     figs["norm"].update_layout(
#         xaxis_title=title_x,
#         yaxis_title=title_y,
#         title=f"{global_title}",
#     )
#     return [figs["norm"]]

# def cryoscope_demod_fft(folder, routine, qubit, format):
#     data = DataUnits.load_data(folder, "data", routine, format, "data")
#     import numpy as np
#     import scipy.signal as ss

#     MX_tag = "MX"
#     MY_tag = "MY"

#     amplitude = data.get_values("flux_pulse_amplitude", "dimensionless")
#     duration = data.get_values("flux_pulse_duration", "ns")
#     flux_pulse_duration = duration[data.df["component"] == MY_tag].to_numpy()
#     flux_pulse_amplitude = amplitude[data.df["component"] == MY_tag].to_numpy()
#     amplitude_unique = flux_pulse_amplitude[
#         flux_pulse_duration == flux_pulse_duration[0]
#     ]
#     duration_unique = flux_pulse_duration[
#         flux_pulse_amplitude == flux_pulse_amplitude[0]
#     ]

#     # Making figure
#     figs = {}
#     figs["norm"] = make_subplots(
#         rows=1,
#         cols=1,
#         horizontal_spacing=0.1,
#         vertical_spacing=0.2,
#     )

#     for amp in amplitude_unique:
#         x_data = data.get_values("prob", "dimensionless")[data.df["flux_pulse_amplitude"] == amp][data.df["component"] == MX_tag].to_numpy()
#         y_data = data.get_values("prob", "dimensionless")[data.df["flux_pulse_amplitude"] == amp][data.df["component"] == MY_tag].to_numpy()

#         complex_data = x_data + 1j * y_data
#         #print (len(complex_data))
#         ws = 10 # Needs to be adjusted for data analysis
#         norm_data = normalize_sincos(complex_data, window_size=ws)

#         sampling_rate = 1 / (duration_unique[1] - duration_unique[0])
#         derivative_window_length = 7 / sampling_rate
#         derivative_window_size = max(3, int(derivative_window_length * sampling_rate))
#         derivative_window_size += (derivative_window_size + 1) % 2
#         derivative_order = 2
#         nyquist_order=0

#         freq_guess, phase_guess, offset_guess, amp_guess = fft_based_freq_guess_complex(norm_data)
#         demod_freq = - \
#                 freq_guess * sampling_rate

#         #y = amp*exp(2pi i f t + ph) + off.
#         #demod_data = amp_guess * np.exp(2 * np.pi * 1j * demod_freq * duration_unique + phase_guess) + offset_guess
#         demod_data = np.exp(2 * np.pi * 1j * duration_unique * demod_freq) * norm_data
#         # Di Carlo has smooth here!!!

#         # <X> sin <Y> cos demod data
#         global_title = "<X> , <Y> after FFT demodulation"
#         title_x = "Flux Pulse duration"
#         title_y = "<X> , <Y>"
#         figs["norm"].add_trace(
#             go.Scatter(
#                 x=duration_unique,
#                 y=demod_data.real,
#                 name=f"<X> fft demod | A = {amp:.3f}"
#             ),
#             row=1,
#             col=1,
#         )

#         figs["norm"].add_trace(
#             go.Scatter(
#                 x=duration_unique,
#                 y=demod_data.imag,
#                 name=f"<Y> fft demod | A = {amp:.3f}"
#             ),
#             row=1,
#             col=1,
#         )

#     figs["norm"].update_layout(
#         xaxis_title=title_x,
#         yaxis_title=title_y,
#         title=f"{global_title}",
#     )
#     return [figs["norm"]]



import plotly.graph_objects as go
from plotly.subplots import make_subplots


def flux_dependence_plot(data, fit, qubit):
    figures = []
    fitting_report = "No fitting data"

    qubit_data = data.df[data.df["qubit"] == qubit]

    fig = make_subplots(
        rows=1,
        cols=2,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        subplot_titles=(
            "MSR [V]",
            "Phase [rad]",
        ),
    )

    fig.add_trace(
        go.Heatmap(
            x=qubit_data["frequency"].pint.to("GHz").pint.magnitude,
            y=qubit_data["bias"].pint.to("V").pint.magnitude,
            z=qubit_data["MSR"].pint.to("uV").pint.magnitude,
            colorbar_x=0.46,
        ),
        row=1,
        col=1,
    )
    fig.update_xaxes(
        title_text=f"{qubit}: Frequency (Hz)",
        row=1,
        col=1,
    )
    fig.update_yaxes(title_text="Bias (V)", row=1, col=1)

    fig.add_trace(
        go.Heatmap(
            x=qubit_data["frequency"].pint.to("GHz").pint.magnitude,
            y=qubit_data["bias"].pint.to("V").pint.magnitude,
            z=qubit_data["phase"].pint.to("rad").pint.magnitude,
            colorbar_x=1.01,
        ),
        row=1,
        col=2,
    )
    fig.update_xaxes(
        title_text=f"{qubit}: Frequency (Hz)",
        row=1,
        col=2,
    )
    fig.update_yaxes(title_text="Bias (V)", row=1, col=2)

    fig.update_layout(
        showlegend=False,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
    )

    figures.append(fig)

    return figures, fitting_report



from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler


def scikit_normalize(constructor):
    r"""Returns a `Pipeline` with `StandardScaler` and the
    `constructor`.

    Args:

        constructor: `sklearn` model.
    """
    return make_pipeline(StandardScaler(), constructor)


def identity(x):
    return x



from dataclasses import dataclass, field
from typing import Dict, List, Tuple

import numpy as np
import plotly.graph_objects as go
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal.auto.operation import Parameters, Qubits, Results, Routine
from qibocal.config import log
from qibocal.plots.utils import get_color

from .t1 import T1Data
from .utils import exp_decay, exponential_fit


@dataclass
class SpinEchoParameters(Parameters):
    """SpinEcho runcard inputs."""

    delay_between_pulses_start: int
    """Initial delay between pulses (ns)."""
    delay_between_pulses_end: int
    """Final delay between pulses (ns)."""
    delay_between_pulses_step: int
    """Step delay between pulses (ns)."""


@dataclass
class SpinEchoResults(Results):
    """SpinEcho outputs."""

    t2_spin_echo: Dict[List[Tuple], str] = field(metadata=dict(update="t2_spin_echo"))
    """T2 echo for each qubit."""
    fitted_parameters: Dict[List[Tuple], List]
    """Raw fitting output."""


class SpinEchoData(T1Data):
    """SpinEcho acquisition outputs."""


def _acquisition(
    params: SpinEchoParameters,
    platform: AbstractPlatform,
    qubits: Qubits,
) -> SpinEchoData:
    """Data acquisition for SpinEcho"""
    # create a sequence of pulses for the experiment:
    # Spin Echo 3 Pulses: RX(pi/2) - wait t(rotates z) - RX(pi) - wait t(rotates z) - RX(pi/2) - readout
    ro_pulses = {}
    RX90_pulses1 = {}
    RX_pulses = {}
    RX90_pulses2 = {}
    sequence = PulseSequence()
    for qubit in qubits:
        RX90_pulses1[qubit] = platform.create_RX90_pulse(qubit, start=0)
        RX_pulses[qubit] = platform.create_RX_pulse(
            qubit, start=RX90_pulses1[qubit].finish
        )
        RX90_pulses2[qubit] = platform.create_RX90_pulse(
            qubit, start=RX_pulses[qubit].finish
        )
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RX90_pulses2[qubit].finish
        )
        sequence.add(RX90_pulses1[qubit])
        sequence.add(RX_pulses[qubit])
        sequence.add(RX90_pulses2[qubit])
        sequence.add(ro_pulses[qubit])

    # define the parameter to sweep and its range:
    # delay between pulses
    ro_wait_range = np.arange(
        params.delay_between_pulses_start,
        params.delay_between_pulses_end,
        params.delay_between_pulses_step,
    )

    data = SpinEchoData()

    # sweep the parameter
    for wait in ro_wait_range:
        # save data as often as defined by points

        for qubit in qubits:
            RX_pulses[qubit].start = RX90_pulses1[qubit].finish + wait
            RX90_pulses2[qubit].start = RX_pulses[qubit].finish + wait
            ro_pulses[qubit].start = RX90_pulses2[qubit].finish

        # execute the pulse sequence
        results = platform.execute_pulse_sequence(sequence)

        for ro_pulse in ro_pulses.values():
            # average msr, phase, i and q over the number of shots defined in the runcard
            r = results[ro_pulse.serial].average.raw
            r.update(
                {
                    "wait[ns]": 2 * wait,
                    "qubit": ro_pulse.qubit,
                }
            )
            data.add(r)
    return data


def _fit(data: SpinEchoData) -> SpinEchoResults:
    """Post-processing for SpinEcho."""
    t2Echos, fitted_parameters = exponential_fit(data)

    return SpinEchoResults(t2Echos, fitted_parameters)


def _plot(data: SpinEchoData, fit: SpinEchoResults, qubit: int):
    """Plotting for SpinEcho"""

    figures = []
    fig = go.Figure()

    # iterate over multiple data folders
    fitting_report = ""

    qubit_data = data.df[data.df["qubit"] == qubit]
    waits = data.df["wait"].pint.to("ns").pint.magnitude

    fig.add_trace(
        go.Scatter(
            x=qubit_data["wait"].pint.to("ns").pint.magnitude,
            y=qubit_data["MSR"].pint.to("uV").pint.magnitude,
            marker_color=get_color(0),
            opacity=1,
            name="Voltage",
            showlegend=True,
            legendgroup="Voltage",
        ),
    )

    # add fitting trace
    if len(data) > 0:
        waitrange = np.linspace(
            min(waits),
            max(waits),
            2 * len(data),
        )
        params = fit.fitted_parameters[qubit]

        fig.add_trace(
            go.Scatter(
                x=waitrange,
                y=exp_decay(waitrange, *params),
                name="Fit",
                line=go.scatter.Line(dash="dot"),
                marker_color=get_color(1),
            ),
        )

        fitting_report = fitting_report + (
            f"{qubit} | T2 Spin Echo: {fit.t2_spin_echo[qubit]:,.0f} ns.<br><br>"
        )

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="Time (ns)",
        yaxis_title="MSR (uV)",
    )

    figures.append(fig)

    return figures, fitting_report


spin_echo = Routine(_acquisition, _fit, _plot)
"""SpinEcho Routine object."""
